import portion as Int
import ply.lex as lex
import ply.yacc as yacc

from .problem import Problem
from .venn import Venn
from .configuration import *
from .level_1 import *
from .level_2 import *


class Lexer(object):
    def __init__(self):
        self.lexer = lex.lex(module=self)

    reserved = {
        "labelled": "LAB",
        "property": "PROP",
        "in": "IN",
        "universe": "UNIVERSE",
        "part": "PART",
        "repeated": "REPEAT",
    }

    # Tokens

    t_EQUALS = r"="
    t_GT = r">"
    t_LT = r"<"
    # t_COL = r":"
    t_SEMI = r";"
    # t_SLASH : r'\?'
    t_COMMA = r","
    t_LPAR = r"\{"
    t_RPAR = r"\}"
    t_LSPAR = r"\["
    t_RSPAR = r"\]"
    t_LRPAR = r"\("
    t_RRPAR = r"\)"
    t_COUNT = r"\#"
    t_UNION = r"\+"
    t_INTER = r"\&"
    t_DIFF = r"!"
    t_NOT = r"\¬"
    t_ignore_COMMENT = r"%.*"
    t_ignore_WHITES = r"\ +|\t|\n"

    def t_LABEL(self, t):
        r"[a-z][a-zA-Z\-\_0-9]*|[0-9]+\-[0-9\-]+"
        t.type = self.reserved.get(t.value, "LABEL")  # Check for reserved words
        return t

    def t_NUMBER(self, t):
        r"\d+"
        try:
            t.value = int(t.value)
        except ValueError:
            print("Integer value too large %d", t.value)
            t.value = 0
        return t

    def t_newline(self, t):
        r"\n+"
        t.lexer.lineno += t.value.count("\n")

    def t_error(self, t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    tokens = [
        "COUNT",
        "COMMA",
        "EQUALS",
        "LT",
        "GT",
        # "COL",
        "SEMI",
        "LPAR",
        "RPAR",
        "LSPAR",
        "RSPAR",
        "LRPAR",
        "RRPAR",
        "NUMBER",
        "UNION",
        "INTER",
        "NOT",
        "DIFF",
        "LABEL",
    ] + list(reserved.values())


class Parser(object):

    tokens = Lexer.tokens

    def __init__(self, file=None, cola=None):
        self.file = file
        self.cola = cola
        self.lexer = Lexer()
        self.parser = yacc.yacc(module=self)
        self.sizes = {}
        self.problem = Problem()
        self.venn = Venn()

    def parse(self):
        if self.file is not None:
            with open(self.file, "r") as f:
                cola = f.read()
        else:
            cola = self.cola
        self.parser.parse(cola)
        if self.problem.configuration is None:
            raise EmptyException("No configuration specified")

    def p_program(self, p):
        """program : statement
        | statement program
        """

    def p_statement(self, p):
        """statement : declare_set SEMI
        | arrangement SEMI
        | pos_constraint SEMI
        | count_constraint SEMI
        """

    def p_entity_list(self, p):
        """entity_list : LABEL
        | LABEL COMMA entity_list
        """
        if len(p) > 2:
            p[0] = [p[1]] + p[3]
        else:
            p[0] = [p[1]]

    def p_comp(self, p):
        """comp : EQUALS
        | LT
        | GT
        | GT EQUALS
        | LT EQUALS
        | DIFF EQUALS
        """
        if len(p) > 2:
            p[0] = p[1] + p[2]
        else:
            p[0] = p[1]

    def p_base_set(self, p):
        """base_set : LABEL
        | LABEL LSPAR NUMBER RSPAR
        | UNIVERSE
        """
        if len(p) == 2:
            p[0] = p[1]
        else:
            # need more elegant way to pass number of lifted set
            p[0] = LiftedSet(self.problem.universe, CSize(p[3], Int.closed(1, Int.inf)))

    def p_set_atom(self, p):
        """set_atom : base_set
        | LRPAR set RRPAR
        """
        if p[1] == "(":
            p[0] = p[2]
        else:
            p[0] = p[1]

    def p_set_not(self, p):
        """set_not : NOT set_atom
        | set_atom
        """
        p[0] = p[1]

    def p_set_inter(self, p):
        """set_inter : set_not
        | set_not INTER set_atom
        """
        if len(p) > 2:
            if isinstance(p[1], LiftedSet):
                d = self.problem.compute_dom(p[3])
                p[1].ccs = [CCounting(d, Int.open(0, Int.inf))]
                p[0] = p[1]
            elif isinstance(p[3], LiftedSet):
                d = self.problem.compute_dom(p[1])
                p[3].ccs = [CCounting(d, Int.open(0, Int.inf))]
                p[0] = p[1]
            else:
                p[0] = And(p[1], p[3])
        else:
            p[0] = p[1]

    def p_set_union(self, p):
        """set_union : set_inter
        | set UNION set_inter
        """
        if len(p) > 2:
            if isinstance(p[1], LiftedSet):
                raise Exception("Cannot or partition")
            else:
                p[0] = Or(p[1], p[3])
        else:
            p[0] = p[1]

    def p_set(self, p):
        """set : set_union
        | PART
        """

        if len(p) == 2:
            p[0] = p[1]
        else:
            if isinstance(p[2], LiftedSet):
                raise Exception("Cannot negate partition")
            else:
                p[0] = Not(p[2])

    def p_declare_set(self, p):
        """declare_set : PROP LABEL EQUALS LPAR entity_list RPAR
        | UNIVERSE LABEL EQUALS LPAR entity_list RPAR
        | PROP UNIVERSE EQUALS LPAR entity_list RPAR
        | LAB PROP LABEL EQUALS LPAR entity_list RPAR
        | PROP LABEL
        | LAB PROP LABEL
        """
        enumerated = "=" in p
        implicit_labelled = p[1] == "labelled"
        explicit_univ = p[1] == "universe"
        distinguishable = enumerated or implicit_labelled
        label = p[3] if implicit_labelled else p[2]
        if "=" not in p:
            self.venn.add_base_set(label, distinguishable)
        else:
            k = 1 if implicit_labelled else 0
            elems = p[5 + k]
            implicit_univ = enumerated and self.problem.universe.empty()
            univ = implicit_univ or explicit_univ
            elems = list2interval(self.problem, elems, univ)
            if explicit_univ:
                d = Universe(label, elems, name=label)
                self.problem.universe = d
            else:
                d = SetFormula(label, elems, self.problem.universe, name=label)
            self.problem.add_domain(d)
            p[0] = d

    def p_arrangement(self, p):
        """arrangement : LABEL IN LPAR REPEAT set RPAR
        | LABEL IN LPAR set RPAR
        | LABEL IN LSPAR REPEAT set RSPAR
        | LABEL IN LSPAR set RSPAR
        | LABEL IN LPAR LPAR set RPAR RPAR
        | LABEL IN LSPAR LPAR set RPAR RSPAR
        """
        # not checking var consistency and set existence
        name = p[1]
        if p[3] == "[":
            if p[4] == "repeated":
                type = "sequence"
            elif p[4] == "{":
                type = "composition"
            else:
                type = "permutation"
        else:  # p[3] == "{"
            if p[4] == "repeated":
                type = "multisubset"
            elif p[4] == "{":
                type = "partition"
            else:
                type = "subset"
        set = p[4] if type in ["subset", "permutation"] else p[5]
        if len(self.venn.base_sets) > 0:
            for set in self.sizes:
                self.venn.add_set(set, self.sizes[set])
            self.problem = self.venn.update_domains(self.problem)
        dom = self.problem.compute_dom(set)
        s = Configuration(name, type, dom)
        self.problem.configuration = s
        if self.problem.universe.empty():
            self.problem.compute_universe()
        if len(self.problem.domains) > 0:
            names = list(self.problem.domains)
            for i, d1 in enumerate(names):
                for d2 in names[i:]:
                    dom1 = self.problem.domains[d1]
                    dom2 = self.problem.domains[d2]
                    if dom1.elements.domain() in dom2.elements.domain():
                        dom1.elements.combine(
                            dom2.elements, how=is_distinguishable
                        )  # update indistinguishable
                    if dom2.elements.domain() in dom1.elements.domain():
                        self.problem.domains[d1] = dom1 | dom2
        else:
            raise EmptyException("No sets found")
        p[0] = s

    def p_pos_constraint(self, p):
        """pos_constraint : LABEL LSPAR NUMBER RSPAR EQUALS set
        | LABEL LSPAR NUMBER RSPAR IN set
        | LABEL LSPAR NUMBER RSPAR EQUALS LPAR entity_list RPAR
        """
        arrangement = p[1]
        pos = p[3]
        if p[6] == "{":
            if p[8] == "}":
                set = list2interval(self.problem, p[7], False)
                df = SetFormula("anon", set, self.problem.universe)
            else:
                size = CSize("", Int.closed(1, self.problem.universe.size() - 1))
                df = LiftedSet(self.problem.universe, size)
                for sc in p[7]:
                    if isinstance(sc.formula.name, str):
                        df.size = CSize("", sc.values)
                    else:
                        df.ccs.append(sc)
        else:
            df = self.problem.compute_dom(p[6])
        pf = CPosition(arrangement, pos, df)
        self.problem.add_pos_formula(pf)
        p[0] = pf

    def p_count_constraint(self, p):
        """count_constraint : COUNT set comp NUMBER
        | COUNT LRPAR count_constraint RRPAR comp NUMBER
        """
        if isinstance(p[2], str) and p[2] == "(":
            # level 2 counting constraint
            cf_par = p[3]
            if cf_par.formula.name == self.problem.universe.name:
                cf_par = CSize(p[5], cf_par.values)
            comp = p[5]
            n = p[6]
            interval = self.problem.get_interval(comp, n)
            cf = CCounting(cf_par, interval)
            self.problem.add_counting_formula(cf)
            p[0] = cf
        else:  # level 1 counting constraint
            set = p[2]
            inter = self.problem.get_interval(p[3], p[4])
            size = CSize(set, inter)
            if isinstance(set, LiftedSet):
                # create a lifted set for each position and later conjoin with
                # variable
                pos = set.size.name  # yep this is ugly
                if len(set.ccs) == 0:
                    set.size = size
                else:
                    set.ccs[0].values = inter
                pf = CPosition(self.problem.configuration, pos, set)
                self.problem.add_pos_formula(pf)
            else:
                conf = self.problem.configuration
                if conf is None:
                    # parsing sizes of sets before config declaration
                    self.venn.add_set(set, p[4])
                elif conf is not None and set == conf.name:
                    if size.values.lower == 0:
                        size.values = size.values.replace(lower=1)
                    self.problem.configuration.size = size
                else:
                    df = self.problem.compute_dom(set)
                    cf = CCounting(df, inter)
                    self.problem.add_counting_formula(cf)
                    p[0] = cf

    def p_error(self, p):
        if p == None:
            token = "end of file"
        else:
            token = f"{p.type}({p.value}) on line {p.lineno}"
            # ignore propery as label
            if p.type != "PROPERTY":
                print(f"Syntax error: Unexpected {token}")



import portion as Int

from .solver import Solver
from .level_1 import SetFormula, Universe
from .configuration import CSize
from .util import *


class EmptyException(Exception):
    def __init__(self, message):
        super().__init__(message)


class Problem(object):
    """
    Represents a problem as a collection of domains, a target configuration, a set of choice formulas and a set of constraint formulas.

    Attributes
    ----------
    universe : str
        the name of the largest domain that includes all the others (assumes it was declared)
    entity_map : dict str->int
        maps each entity/constant to an integer (manipulated as intervals)
    """

    def __init__(self):
        self.universe = Universe("", IntervalDict())
        # self.agg_formulas = []
        self.pos_constraints = []
        self.constraints = []
        self.domains = {}
        self.entity_map = {}
        self.configuration = None
        self.internal_copies = {}

    # def add_choice_formula(self, head, body):
    #     struct_name = head.args[0]
    #     formula = head.args[-1]
    #     df = self.compute_dom(formula)
    #     if head.functor == "pos":
    #         pos = head.args[1]
    #         c = PosFormula(struct_name, pos, df)
    #     if head.functor == "in":
    #         c = InFormula(self.configuration.name, df)
    #     self.choice_formulas.append(c)

    def add_domain(self, dom):
        self.domains[dom.name] = dom

    def add_pos_formula(self, chf):
        self.pos_constraints.append(chf)

    def add_agg_formula(self, chf):
        self.agg_formulas.append(chf)

    def add_counting_formula(self, cof):
        # cformula = self.build_cof(cof)
        subsets = self.configuration.type in ["partition", "composition"]
        count_prop = isinstance(cof.formula, SetFormula)
        # ignore recursive parsing of counting formulas
        if not (subsets and count_prop):
            self.constraints.append(cof)

    def add_entity(self, e):
        """
        Map an entity e to its internal representation (id)

        Args:
            e (str): the label of an entity

        Returns:
            int: the id representing the entity internally
        """
        name = str(e)
        if name in self.entity_map:
            return self.entity_map[name]
        else:
            i = len(self.entity_map) + 1
            self.entity_map[name] = i
            copy_ids = name.split("__")
            if len(copy_ids) > 1:
                base_id, _ = copy_ids
                if base_id in self.internal_copies:
                    self.internal_copies[base_id] += [i]
                else:
                    self.internal_copies[base_id] = [i]
            return i

    def get_entity(self, e):
        name = str(e)
        base_id = self.entity_map.get(name)
        if base_id is None:
            raise Exception(f"Unknown label {name}")
        copy_ids = self.internal_copies.get(name, [])
        return [base_id] + copy_ids

    # def add_query(self, q):
    #     self.queries.append(q)

    # def add_size(self, sf):
    #     name = sf.args[0]
    #     op = sf.args[1]
    #     formula = sf.args[2]
    #     s = self.build_size(name,op,formula)
    #     self.configuration.size = s

    # def add_configuration(self, struct):
    #     self.configuration = struct

    # def build_cof(self, cof):
    #     struct_name = cof.args[0]
    #     problog_formula = cof.args[1]
    #     op = cof.args[2].functor
    #     val = cof.args[3].compute_value()
    #     formula = self.compute_formula(problog_formula)
    #     interval = self.get_interval(op, val)
    #     return CountingFormula(formula, interval)

    # def build_size(self, name, op, val):
    #     n = val.compute_value()
    #     op = op.functor
    #     interval = self.get_interval(op,n)
    #     return SizeFormula(name, interval)

    def compute_dom(self, sformula):
        """
        Given a set-formula expands and computes the corresponding domain
        Parameters
        ----------
        sformula : a set formula : an And/Or/Not (possibly nested)
                  or a string corresponding to one of the domains
        """
        if isinstance(sformula, And):
            lf = self.compute_dom(sformula.left)
            rf = self.compute_dom(sformula.right)
            domain = lf & rf
        elif isinstance(sformula, Or):
            lf = self.compute_dom(sformula.left)
            rf = self.compute_dom(sformula.right)
            domain = lf | rf
        elif isinstance(sformula, Not):
            arg = self.compute_dom(sformula.child)
            domain = arg.neg()
        else:  # base cases: universe, arrangement, user-defined, single element
            if sformula in self.domains:
                domain = self.domains[sformula]
            elif (
                sformula == "universe"
                or sformula == "part"
                or self.configuration is not None
                and sformula == self.configuration.name
            ):
                domain = self.universe
            else:
                id = self.get_entity(sformula)
                if id is None:
                    id = ""  # sformula # dummy
                    single = Int.closed(0, P.inf)
                    # raise Exception(f"Unknown constant {sformula}")
                else:
                    single = Int.closed(id[0], id[-1])
                    # single = Int.singleton(id)
                dist = Int.IntervalDict()
                dist[single] = True
                domain = SetFormula(id, dist, self.universe)
        return domain

    def compute_universe(self):
        """
        Once all sets are added, compute the union and update the domain formulas
        with all info about the universe
        """
        dom_iter = iter(self.domains.values())
        for d in dom_iter:
            self.universe = self.universe | d
        all_elem = self.universe.elements
        self.universe = Universe(
            self.universe.formula, all_elem, self.universe.get_label(support(all_elem))
        )  # promote set formula
        dom_iter = iter(self.domains.values())
        for d in dom_iter:
            d.universe = self.universe
        for pf in self.pos_constraints:
            pf.universe = self.universe
        for cof in self.constraints:
            cof.universe = self.universe
        self.set_labels()
        # cof.formula.universe = universe

    # def compute_formula(self, formula):
    #     if formula.functor == "size":
    #         name = formula.args[0]
    #         op = formula.args[1]
    #         n = formula.args[2]
    #         return self.build_size(name,op,n)
    #     elif formula.functor == "count":
    #         return self.build_cof(formula)
    #     else:
    #         return self.compute_dom(formula)

    # def get_entity(self, e):
    #     if e in self.entity_map:
    #         return self.entity_map[e]
    #     else:
    #         return None

    def get_interval(self, op, n):
        """
        Convert a comparison operator and a number to interval

        Args:
            op (str): a string describing a comparison
            n (int): reference value

        Returns:
            Int: interval of values
        """
        if op == "<":
            interval = Int.closedopen(0, n)
        elif op == "<=":
            interval = Int.closed(0, n)
        elif op == ">":
            interval = Int.open(n, Int.inf)
        elif op == ">=":
            interval = Int.closedopen(n, Int.inf)
        elif op == "=":
            interval = Int.singleton(n)
        else:
            interval = Int.closedopen(0, n) | Int.open(n, Int.inf)
        return interval

    def set_labels(self):
        """
        Adds to the universe the original labels of the elements
        """
        self.universe.labels_entity = {
            v: self.trim(k) for k, v in self.entity_map.items()
        }
        for k, v in self.domains.items():
            dom = P.empty()
            for d in v.elements.keys():
                dom = dom | d
            self.universe.add_set_label(dom, k)

    def solve(self, debug=True):
        """
        Do some sanity checks on the configuration and then call the solver

        Args:
            log (bool, optional): enable logging. Defaults to True.

        Returns:
            Solution: solution
        """
        if self.configuration is None or len(self.domains) == 0:
            raise EmptyException("Empty problem!")
        else:
            if self.configuration.size is None:
                vals = Int.closed(1, self.universe.size())
                self.configuration.size = CSize("unconstrained", vals)
            self.set_labels()
            s = Solver(self, debug=debug)
            return s.solve()

    def trim(self, entity):
        """
        Removes the internal copy id from an indistinguishable entity's name

        Args:
            entity (str): label of the entity

        Returns:
            str: the label if entity is distinguishable, the base lable otherwise
        """
        e = entity.split("__")
        return e[0]

    def __str__(self):
        s = ""
        for d in self.domains.values():
            s += f"{d.name}: {d}\n"
        for cf in self.pos_constraints:
            s += f"{cf}\n"
        for f in self.constraints:
            s += f"{f}\n"
        # for f in self.agg_formulas:
        #     s += f"{f}\n"
        s += f"{self.configuration}\n"
        return s



import portion as P

from .configuration import Variable
from .util import *


class Multiset(object):
    """
    Represents a domain as a portion (set of intervals) with extra info, i.e. distinguishability, size
    and take operation

    Attributes
    ----------
    name : str
        the label associated to the multiset
    formula: str
        the set-formula describing the multiset (usually same as name)
    elements : IntervalDict
        Intervals corresponding to the entities (mapped to integers) of the multiset.
        Mapped as dict to True if distinguishable False otherwise
    """

    def __init__(self, name, elems):
        self.name = name
        self.formula = name
        self.elements = elems
        self.n_elements = None

    def __contains__(self, val):
        return val.elements.keys() <= self.elements.keys()

    def __eq__(self, rhs):
        return self.elements == rhs.elements

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return str(self)

    def __str__(self):
        if self.name is not None:
            return str(self.name)
        else:
            s = f"{self.formula}"
            if self.all_indistinguishable() and len(self.elements.domain()) == 1:
                n_elems = sum([1 for _ in P.iterate(self.elements.domain(), step=1)])
                i = self.elements.domain().lower
                # e_lab = self.labels.get(i, i)
                s = f"{n_elems}x " + self.name  # str(e_lab)
        # ok for debug, not for log
        # if self.size() > 0 :
        #     s += f"({list(self.elements.keys())})=|{self.n_elements}|"
        # else:
        #     s += f"(none)"
        return s

    def all_indistinguishable(self):
        return not (True in self.elements.values())

    def max(self):
        return self.elements.domain().upper

    def min(self):
        return self.elements.domain().lower

    def disjoint(self, rhs):
        inter = self & rhs
        return inter.elements.domain().empty

    # TODO @property
    def size(self):
        if self.n_elements is None:
            s = 0
            for e in self.elements.domain():
                if e.upper == P.inf:
                    return -1
                if not e.empty:
                    if e.left == P.CLOSED and e.right == P.CLOSED:
                        s += e.upper - e.lower + 1
                    elif e.left == P.OPEN and e.right == P.OPEN:
                        s += e.upper - e.lower - 1
                    else:
                        s += e.upper - e.lower
            self.n_elements = s
        return self.n_elements

    def take(self, n):
        """
        returns a subset of itself of size n
        if n > size returns itself
        since we use it for shattering alldiff, assumes that all elements are distinguishable
        """
        iter = P.iterate(self.elements.domain(), step=1)
        subset = P.empty()
        i = 0
        hasNext = True
        while hasNext and i < n:
            try:
                elem = P.singleton(next(iter))
                subset = subset | elem
            except StopIteration:
                hasNext = False
            i += 1
        dist = P.IntervalDict()
        dist[subset] = True
        taken = Multiset(f"{n}x {self.name}", dist)
        return taken


class SetFormula(Multiset, Variable):
    """
    Represents a set by means of intersections/union/complements of the declared sets

    Attributes
    ----------
    name : str/And/Or/Not
        the description of the set operations to generate the set
    elems : DictInterval
        the corresponding elements
    universe : Multiset
        the universe of the problem: each complement is computed w.r.t. it
    """

    def __init__(self, formula, elems, universe, name=None):
        Variable.__init__(self, universe, name)
        Multiset.__init__(self, formula, elems)
        if name is None:
            self.name = self.universe.get_label(support(elems), str(formula))
        else:
            self.name = name

    def __add__(self, rhs):
        return self.elements.combine(rhs.elements, how=is_distinguishable)

    def __and__(self, rhs):
        """
        Intersection of SetFormulas self /\ rhs

        Args:
            rhs (SetFormulas): another SetFormulas

        Returns:
            SetFormulas: SetFormulas representing the intersection
        """
        dom = self.elements.domain() & rhs.elements.domain()
        # comb = self.elements.combine(rhs.elements, how=SetFormula.is_distinguishable)
        comb = combine(self, rhs)
        elems = comb[dom]
        if elems == self.elements:
            return self
        elif elems == rhs.elements:
            return rhs
        else:
            f = And(self, rhs)
            default = simplify_name(elems, str(f), self.universe)
            name = self.universe.get_label(dom, default)
            return SetFormula(f, elems, self.universe, name)

    def __or__(self, rhs):
        """
        Disjunction of SetFormulas

        Args:
            rhs (SetFormulas): other SetFormulas

        Returns:
            SetFormulas: union of self with rhs
        """
        comb = combine(self, rhs)
        if comb.domain() in self.elements.domain():
            return self
        elif comb.domain() in rhs.elements.domain():
            return rhs
        else:
            f = Or(self, rhs)
            default = simplify_name(comb, str(f), self.universe)
            name = self.universe.get_label(support(comb), default)
            return SetFormula(f, comb, self.universe, name)

    def __sub__(self, rhs):
        """
        Remove objects in rhs from self

        Args:
            rhs (SetFormulas): objects to remove

        Returns:
            SetFormulas: self-rhs
        """
        if self.disjoint(rhs):
            f = self.formula
        else:
            f = And(self, Not(rhs))
        comb = self.elements.combine(rhs.elements, how=is_distinguishable)
        sub_dom = self.elements.domain() - rhs.elements.domain()
        return SetFormula(f, comb[sub_dom], self.universe, self.name)

    def __iter__(self):
        for e in P.iterate(support(self.elements), step=1):
            yield e

    def copy(self):
        return SetFormula(self.formula, self.elements, self.universe, self.name)

    def get_label(self, key, default=None):
        return self.universe.get_label(key, default)

    def empty(self):
        return support(self.elements).empty

    def neg(self):
        """
        Complement the set w.r.t. universe

        Returns:
            SetFormulas: universe-self
        """
        els = self.universe.elements.domain() - self.elements.domain()
        dom = self.universe.elements[els]
        if isinstance(self.formula, Not):
            f = self.formula.child
        else:
            f = Not(self)
        name = str(Not(str(self)))
        if self.universe is not None:
            name = self.universe.get_label(els, name)
        return SetFormula(f, dom, self.universe, name)

    def indistinguishable_subsets(self, set_formula=None):
        """
        Collect all partitions of indistinguishable objects contained in the set formula

        Args:
            set_formula (SetFormula, optional): set formula . Defaults to None.

        Returns:
            [portion]: intervals corresponding to the ids of indistinguishable objects
        """
        df = set_formula if set_formula is not None else self
        f = df.formula
        if isinstance(f, And) or isinstance(f, Or):
            ind_l = self.indistinguishable_subsets(f.left)
            ind_r = self.indistinguishable_subsets(f.right)
            indist = ind_l.union(ind_r)
        elif isinstance(f, Not):
            indist = self.indistinguishable_subsets(f.child)
        else:  # Domain
            df = df & self.universe  # update indistinguishability from universe
            vals = list(df.elements.values())
            if [False] == vals:
                indist = {df}
            elif False in vals:
                indist_subsets = set()
                i = 1
                for dom in df.elements.find(False):
                    sub = SetFormula(f"{df.formula}_{i}", df.elements[dom], df.universe)
                    indist_subsets.add(sub)
                    i += 1
                return indist_subsets
            else:
                indist = set()
        return indist

    @Variable.universe.setter
    def universe(self, universe):
        """
        Propagate universe info to subsets

        Args:
            universe (Domain): universe
        """
        self._universe = universe
        if isinstance(self.formula, And) or isinstance(self.formula, Or):
            if isinstance(self.formula.left, SetFormula):
                self.formula.left.universe = universe
            if isinstance(self.formula.right, SetFormula):
                self.formula.right.universe = universe
        elif isinstance(self.formula, Not):
            if isinstance(self.formula.child, SetFormula):
                self.formula.child.universe = universe
        else:
            pass

    def enumerate_elements(self):
        """
        Enumerates the labels corresponding to the multiset

        Returns:
            str: string representation of the multiset
        """
        labels = [f"{self.get_label(e)}" for e in self]
        avg_length = sum([len(lab) for lab in labels]) / len(labels)
        separator = ",\n\t" if avg_length > 7 else ", "
        lab_list = separator.join(labels)
        if avg_length > 7:
            enum = f"{{\n\t{lab_list}\n}}"
        else:
            enum = f"{{{lab_list}}}"
        return enum

    # def take(self, n):
    #     taken = Multiset.take(n)
    #     return SetFormula(
    #         self.formula,
    #         taken.elements,
    #         self.universe,
    #         name=taken.name,
    #         labels=self.labels,
    #     )


class Universe(SetFormula):
    """
    A SetFormula aware of being the Universe. Keeps track of the user-defined labels associated to individual entities and sets
    """

    def __init__(
        self, formula, elems, name="universe", labels_entity={}, labels_set={}
    ):
        super().__init__(formula, elems, None, name)
        self.labels_entity = labels_entity
        self.labels_set = labels_set
        if len(support(elems)) > 0:
            self.labels_set[support(elems)] = self.name

    def add_set_label(self, domain, label):
        self.labels_set[domain] = label
        complement = support(self.elements) - domain
        if complement not in self.labels_set:
            self.labels_set[complement] = str(Not(label))

    def get_label(self, key, default="???"):
        if isinstance(key, int):
            return self.labels_entity.get(key, default)
        elif isinstance(key, P.Interval):
            if key.empty:
                return "empty"
            self.labels_set.setdefault(key, default)
            return self.labels_set[key]
        else:
            raise Exception(f"Unknown key type for label: {type(key)}")

    def get_labels(self):
        return self.labels_entity, self.labels_set

    # def find_label(self, key):
    #     """Approximates a smart way to get short labels for parts of the universe combining known labels
    #         But it is not guaranteed that the lable is the best possible

    #     Args:
    #         key (Interval): the interval(s) corresponding to the elements of the part

    #     Returns:
    #         str: a human-readable label for the given set
    #     """
    # parents = []
    # for domain in self.labels_set.keys():
    #     if domain.contains(key):
    #         parents.append(domain)
    # if len(parents) == 0:
    #     children = []
    #     for domain in self.labels_set.keys():
    #         if key.contains(domain):
    #             children.append(domain)
    # min_parent = P.empty()
    # for p in parents:
    #     if len(p) < len(min_parent):
    #         min_parent = p
    # print(min_parent, key)
    # for domain in self.labels_set.keys():
    #     if min_parent - domain == key:
    #         lab_mp = self.labels_set[min_parent]
    #         lab_dom = self.labels_set[domain]
    #         lab_dom = lab_dom[1:] if lab_dom.startswith("¬") else f"¬{lab_dom}"
    #         lab_key = f"{lab_mp} ∧ {lab_dom}"
    #         self.labels_entity
    #         return lab_key
    # enum = [self.get_label(k) for k in P.iterate(key, step=1)]
    # lab_list = ", ".join(enum)
    # return f"{{{lab_list}}}"

    def copy(self):
        return Universe(self.formula, self.elements, name=self.name)

    @property
    def universe(self):
        return self



import seaborn as sns
import os
import pandas
import matplotlib
import matplotlib.pyplot as plt

plt.style.use("science")

from .util import ROOT_DIR

OUT_DIR = os.path.join(ROOT_DIR, "tests", "results")
BENCHMARKS = os.path.join(ROOT_DIR, "tests", "benchmarks", "synthetic")
COSO_STATS_BENCH = os.path.join(OUT_DIR, "subs_results_benchmarks.csv")
COSO_STATS_EX = os.path.join(OUT_DIR, "subs_results_examples.csv")
WIDTH = 430
PGF = False

# pal = sns.color_palette("colorblind")


def autolabel(rects, fmt=".2f"):
    # attach some text labels
    for rect in rects:
        height = int(rect.get_height())
        rect.axes.annotate(
            f"{height}",
            xy=(rect.get_x() + rect.get_width() / 2.0, height),
            xytext=(0, 3),
            textcoords="offset points",
            ha="center",
            va="bottom",
            zorder=10,
        )


def set_size(width_pt, fraction=1, subplots=(1, 1)):
    """Set figure dimensions to sit nicely in our document.

    Parameters
    ----------
    width_pt: float
            Document width in points
    fraction: float, optional
            Fraction of the width which you wish the figure to occupy
    subplots: array-like, optional
            The number of rows and columns of subplots.
    Returns
    -------
    fig_dim: tuple
            Dimensions of figure in inches
    """
    # Width of figure (in pts)
    fig_width_pt = width_pt * fraction
    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Golden ratio to set aesthetic figure height
    golden_ratio = (5**0.5 - 1) / 2

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt
    # Figure height in inches
    fig_height_in = fig_width_in * golden_ratio * (subplots[0] / subplots[1])

    return (fig_width_in, fig_height_in)


########################
## Sat problems stats ##
########################


def plot_benchmarks_sat(df, fig, axis, save):

    sns.barplot(
        ax=axis,
        x="n_solutions",
        y="time",
        hue="solver",
        data=df,
        # palette=pal,
        log=True,
    )
    axis.set_axisbelow(True)
    axis.grid(b=True, which="major", axis="y", linestyle="-")

    axis.set_yscale("log", nonpositive="clip")
    axis.set_ylabel("Seconds (log)", fontsize=10)
    axis.set_xlabel("\\# solutions", fontsize=10)

    axis.axhline(y=300, color="r", linestyle="--")

    ticks = [10**x for x in range(-2, 3)] + [300]
    axis.set_yticks(ticks)
    axis.set_yticklabels(ticks, va="center")
    axis.yaxis.set_ticks([], minor=True)
    for tick in axis.xaxis.get_major_ticks():
        tick.tick1line.set_visible(False)
        tick.tick2line.set_visible(False)
        tick.label1.set_visible(False)
        tick.label2.set_visible(False)
    axis.xaxis.tick_bottom()
    ticks = []
    for l in sorted(df["n_solutions"].unique()):
        if l <= 10000:
            ticks.append(str(l))
        else:
            ticks.append(("{:.2e}".format(l)))
    axis.set_xticklabels(ticks)
    axis.tick_params(axis="x", labelrotation=90)
    axis.tick_params(
        axis="x",  # changes apply to the x-axis
        which="both",  # both major and minor ticks are affected
        bottom=False,  # ticks along the bottom edge are off
        top=False,  # ticks along the top edge are off
    )
    handles, labels = axis.get_legend_handles_labels()
    # ph = [plt.plot([], marker="", ls="")[0]]
    # handles = ph + ph + handles
    axis.legend(
        handles,
        ["CoSo", "ASP", "sharpSAT", "Essence"],
        bbox_to_anchor=(0.5, 1.2),
        loc="upper center",
        ncol=4,
        frameon=True,
        title="Framework",
        fontsize=8,
        title_fontsize=10,
        borderaxespad=0,
    )

    axis.set_title("Satisfiable problems", fontsize="10", x=0.5, y=1.22)
    axis.tick_params(axis="both", labelsize=8)

    # plt.subplots_adjust(bottom=0.35, left=0.1)

    if save:
        w, h = set_size(WIDTH)
        fig.set_size_inches(w + 0.1, h)
        path = os.path.join(OUT_DIR, "sat.pgf")
        plt.savefig(path, bbox_inches="tight")
        plt.close(fig)


##########################
## Unsat problems stats ##
##########################


def plot_benchmarks_unsat(df, fig, axis, save):

    sns.barplot(
        ax=axis,
        x="benchmark",
        y="time",
        hue="solver",
        data=df,
        # palette=pal
    )
    axis.set_axisbelow(True)
    axis.grid(b=True, which="major", axis="y", linestyle="-")

    axis.set_yscale("log", nonpositive="clip")
    axis.set_ylabel("Seconds (log)", fontsize="10")
    axis.set_xlabel("Benchmark", fontsize="10")

    axis.axhline(y=300, color="r", linestyle="--")

    ticks = [10**x for x in range(-2, 3)] + [300]
    axis.set_yticks(ticks)
    axis.set_yticklabels(ticks, va="center")
    for tick in axis.xaxis.get_major_ticks():
        tick.tick1line.set_visible(False)
        tick.tick2line.set_visible(False)
        tick.label1.set_visible(False)
        tick.label2.set_visible(False)
    axis.xaxis.tick_bottom()
    # xlabs = [f"unsat_{i}" for i, x in enumerate(axis.get_xticklabels()) ]
    xlabs = [
        "cp 15/5",
        "cp 20/10",
        "ms 20/15",
        "pm 20/15",
        "sq 15/10",
        "sq 20/15",
    ]
    # xlabs = df["benchmark"].unique()
    axis.set_xticklabels(xlabs, va="top")
    axis.tick_params(axis="x", which="major", pad=5)
    axis.tick_params(axis="both", labelsize=8)
    handles, labels = axis.get_legend_handles_labels()
    lgd = axis.legend(
        handles,
        ["CoSo", "ASP", "sharpSAT", "Essence"],
        bbox_to_anchor=(0.5, 1.2),
        loc="upper center",
        ncol=4,
        frameon=True,
        title="Framework",
        fontsize=8,
        title_fontsize=10,
        borderaxespad=0,
    )
    # axis.legend([],[], frameon=False)
    axis.set_title("Unsatisfiable problems", fontsize=10, x=0.5, y=1.22)
    axis.yaxis.set_ticks([], minor=True)
    axis.tick_params(
        axis="x",  # changes apply to the x-axis
        which="both",  # both major and minor ticks are affected
        bottom=False,  # ticks along the bottom edge are off
        top=False,  # ticks along the top edge are off
    )

    plt.subplots_adjust(left=0.15)

    if save:
        w, h = set_size(WIDTH)
        fig.set_size_inches(w, h)
        path = os.path.join(OUT_DIR, "unsat.pgf")
        plt.savefig(path, bbox_extra_artists=(lgd,), bbox_inches="tight")
        plt.close(fig)


############################
## CoSo subproblems stats ##
############################


def plot_coso_stats(df, fig, axis, save):

    count = df.groupby(["n_subproblems", "origin"])["count"].sum().unstack(fill_value=0)
    time = df.groupby(["n_subproblems"], as_index=False)["time"].mean()

    color_cycle = axis._get_lines.prop_cycler
    color_line = next(color_cycle)["color"]
    color_real = next(color_cycle)["color"]
    color_synth = next(color_cycle)["color"]

    count.plot(
        kind="bar",
        stacked=True,
        # palette=[color_real, color_synth],
        color=sns.color_palette([color_real, color_synth]),
        width=0.9,
        ax=axis,
    ).legend(loc="upper center", bbox_to_anchor=(0.5, 1))
    plt.xlim(-1, None)

    n = len(count)
    for i, patch in enumerate(axis.patches):
        c = 0 if i < n else 1
        r = i if i < n else i - n
        x, y = patch.get_xy()
        x += patch.get_width() / 2
        y += patch.get_height() / 2
        l = "" if count.iloc[r, c] == 0 else count.iloc[r, c]
        axis.annotate(l, (x, y), ha="center", va="center", c="black", fontsize=10)

    axis.tick_params(
        axis="x",  # changes apply to the x-axis
        which="both",  # both major and minor ticks are affected
        bottom=False,  # ticks along the bottom edge are off
        top=False,  # ticks along the top edge are off
    )
    ax6 = axis.twinx()
    ax7 = axis.twinx()
    axis.tick_params(axis="both", labelsize=8)
    ax6.tick_params(axis="both", labelsize=8)
    ax6.plot(
        axis.get_xticks(),
        time["time"],
        color=color_line
        # color=sns.color_palette("colorblind")[2]
    )
    axis.set_yscale("symlog", linthresh=10)
    ax6.set_yscale("log", nonpositive="clip")
    ax7.set_yscale("log", nonpositive="clip")

    ax7.set_ylim(ax6.get_ylim())

    axis.grid(False)
    ax6.grid(False)
    ax7.grid(False)
    axis.yaxis.set_ticks([], minor=True)
    ax6.yaxis.set_ticks([], minor=True)
    ax7.yaxis.set_ticks([], minor=True)
    ticks = [10**x for x in range(0, 3)]
    axis.set_yticks(ticks)
    axis.set_yticklabels(ticks, va="center")
    ticks = [10**x for x in range(-1, 3)]
    ax6.set_yticks(ticks)
    ax6.set_yticklabels(ticks, va="center")
    ax7.set(yticks=ax6.get_yticks(), yticklabels=[], ylim=ax6.get_ylim())

    for y in ax6.get_yticks():
        ax7.axhline(y=y, color=ax6.get_lines()[-1].get_c(), alpha=0.3, linestyle="-")

    axis.set_xlabel("\\# subproblems", fontsize=10)
    axis.set_ylabel("\\# benchmarks (log)", fontsize=10)
    ax6.set_ylabel("Seconds (log)", fontsize=10)
    axis.set_title("CoSo avg. runtime vs. \\#subproblems", fontsize=10)

    axis.set_zorder(ax7.get_zorder() + 1)
    ax6.set_zorder(ax7.get_zorder() + 1)
    axis.set_frame_on(False)
    ax6.set_frame_on(False)

    ax7.spines["bottom"].set_color("0.5")
    ax7.spines["top"].set_color("0.5")
    ax7.spines["right"].set_color("0.5")
    ax7.spines["left"].set_color("0.5")

    axis.legend(loc=9, prop={"size": 10})
    plt.subplots_adjust(bottom=0.15, left=0.1)

    if save:
        w, h = set_size(WIDTH)
        fig.set_size_inches(w, h)
        path = os.path.join(OUT_DIR, "subproblems.pgf")
        plt.savefig(path, bbox_inches="tight")
        plt.close(fig)


#####################
## Growing domains ##
#####################


def plot_growing_doms(df, fig, axis, name):

    for ind in df.index:
        s = df["benchmark"][ind]
        df["benchmark"][ind] = 0
        for i in range(1, 11):
            if f"_{i}" in s:
                df["benchmark"][ind] = i

    sns.lineplot(
        ax=axis,
        x="benchmark",
        y="time",
        hue="solver",
        data=df,
        marker=".",
        # palette=pal,
    )

    axis.set_axisbelow(True)
    axis.grid(b=True, which="major", axis="y", linestyle="-")
    axis.tick_params(
        axis="x",  # changes apply to the x-axis
        which="minor",  # both major and minor ticks are affected
        bottom=False,  # ticks along the bottom edge are off
        top=False,  # ticks along the top edge are off
    )
    axis.yaxis.set_ticks([], minor=True)

    # axis.set_yscale("symlog", linthresh=10)

    axis.axhline(y=300, color="r", linestyle="--")
    axis.set_xlabel(name, fontsize=10)

    if name == "P3":
        axis.set_ylabel("Seconds", fontsize=10)
        ticks = [1, 10, 50, 100, 200, 300]
        axis.set_yticks(ticks)
        axis.set_yticklabels(ticks, va="center", fontsize=10)
    axis.tick_params(axis="both", labelsize=8)
    # if name == "P5":
    #     handles, labels = axis.get_legend_handles_labels()
    #     fig.legend(
    #         handles,
    #         ["CoSo", "ASP", "sharpSAT", "Essence"],
    #         bbox_to_anchor=(0.2, -0.5),
    #         loc="lower center",
    #         ncol=2,
    #         frameon=True,
    #         title="Framework",
    #         fontsize="x-small",
    #         title_fontsize="small",
    #         borderaxespad=0,
    #     )
    # else:
    axis.get_legend().remove()


def load_folder(folder):
    b_name = f"bench_results_{folder}.csv"
    b_path = os.path.join(OUT_DIR, b_name)
    s_name = f"subs_results_{folder}.csv"
    s_path = os.path.join(OUT_DIR, s_name)
    if os.path.exists(b_path):
        df_bench = pandas.read_csv(b_path, sep="\t")
        df_subs = pandas.read_csv(s_path, sep="\t")
        for r1 in df_bench.index:
            for r2 in df_bench.index:
                if df_bench["benchmark"][r1] == df_bench["benchmark"][r2]:
                    if df_bench["n_subproblems"][r1] == -1:
                        if df_bench["n_subproblems"][r2] != -1:
                            df_bench["n_subproblems"][r1] = df_bench["n_subproblems"][
                                r2
                            ]
                    if df_bench["n_solutions"][r1] == -1:
                        if df_bench["n_solutions"][r2] != -1:
                            df_bench["n_solutions"][r1] = df_bench["n_solutions"][r2]
        return (df_bench, df_subs)
    else:
        p = pandas.DataFrame()
        return (p, p)


def load_data(bench_dir):

    # examples_dir = os.path.join(bench_dir, "examples")
    df_bench_real, df_coso_real = load_folder("examples")
    df_coso_synth = pandas.DataFrame()
    df_bench_synth = pandas.DataFrame()

    for folder in os.listdir(bench_dir):
        if folder != "examples":
            df_bench_folder, df_subs_folder = load_folder(folder)
            df_bench_synth = pandas.concat([df_bench_synth, df_bench_folder])
            df_coso_synth = pandas.concat([df_coso_synth, df_subs_folder])

    df_sat = df_bench_synth[df_bench_synth["n_solutions"] > 0]
    df_unsat = df_bench_synth[df_bench_synth["n_solutions"] == 0]
    # df_unknown = df_bench_synth[df_bench_synth["n_solutions"] == -1]

    df_coso_synth["origin"] = "synthetic"
    df_coso_real["origin"] = "real"
    df_coso = pandas.concat([df_coso_synth, df_coso_real])
    df_coso = df_coso[df_coso["n_subproblems"] >= 0]

    df_bench_growing = {}
    df_bench_growing_all, _ = load_folder("growing_domains")
    df_bench_growing[1] = df_bench_growing_all[
        df_bench_growing_all["benchmark"].str.startswith("h730")
    ]
    df_bench_growing[2] = df_bench_growing_all[
        df_bench_growing_all["benchmark"].str.startswith("m722")
    ]
    df_bench_growing[3] = df_bench_growing_all[
        df_bench_growing_all["benchmark"].str.startswith("m617")
    ]

    return (df_sat, df_unsat, df_coso, df_bench_real, df_bench_growing)


def plot(bench_dir, pgf=False):
    # sns.set_theme(style="whitegrid", color_codes=True, palette=pal)
    df_sat, df_unsat, df_coso, df_real, df_growing = load_data(bench_dir)

    if pgf:
        matplotlib.use("pgf")
        matplotlib.rcParams.update(
            {
                "pgf.texsystem": "pdflatex",
                "font.family": "serif",
                "text.usetex": True,
                "pgf.rcfonts": False,
            }
        )
    if pgf:
        fig1, ax1 = plt.subplots()
        plot_benchmarks_sat(df_sat, fig1, ax1, pgf)
        fig2, ax2 = plt.subplots()
        plot_benchmarks_unsat(df_unsat, fig2, ax2, pgf)
        fig3, ax3 = plt.subplots()
        plot_coso_stats(df_coso, fig3, ax3, pgf)
    else:
        fig, axes = plt.subplots(nrows=1, ncols=3)
        plot_benchmarks_sat(df_sat, fig, axes[0], pgf)
        plot_benchmarks_unsat(df_unsat, fig, axes[1], pgf)
        plot_coso_stats(df_coso, fig, axes[2], pgf)
    fig_g, axes_g = plt.subplots(nrows=1, ncols=3, sharey=True)
    plot_growing_doms(df_growing[1], fig_g, axes_g[0], "P3")
    plot_growing_doms(df_growing[2], fig_g, axes_g[1], "P4")
    plot_growing_doms(df_growing[3], fig_g, axes_g[2], "P5")
    handles, labels = axes_g[2].get_legend_handles_labels()
    # ph = [plt.plot([], marker="", ls="")[0]]
    # handles = ph + handles
    # labels = ["Framework:"] + labels
    fig_g.legend(
        handles,
        ["CoSo", "ASP", "sharpSAT", "Essence"],
        bbox_to_anchor=(0.5, 1.0),
        loc="upper center",
        ncol=4,
        frameon=True,
        title="Framework",
        fontsize=8,
        title_fontsize=10,
        borderaxespad=0,
    )
    fig_g.tight_layout(pad=0.5)
    if pgf:
        w, h = set_size(WIDTH)
        fig_g.set_size_inches(w, h)
        path = os.path.join(OUT_DIR, "growing_domains.pgf")
        plt.savefig(path, bbox_inches="tight")
    else:
        plt.show()


if __name__ == "__main__":
    plot("")



import portion as P
import os
import sys
from portion.dict import IntervalDict
from typing import Counter

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), ".."))


######################
## Interval methods ##
######################


def is_distinguishable(d1, d2):
    # if e is distinguishable truth value in one domain is d1 and the other d2,
    # if any of the two elements is distinguishable then keep distinguishing
    return d1 or d2


def support(idict):
    """Returns the domain of an intervaldict

    Args:
        idict (IntervalDict): intervals
    """
    support = P.empty()
    for interval in idict.keys():
        support = support.union(interval)
    return support


def interval_closed(interval, lb_default=0, ub_default=sys.maxsize):
    """
    Convert any interval into an equivalent pair (lower bound, upper bound)
    with extremes included

    Args:
        interval (portion): the interval
        lb_default (int, optional): if the lower bound is infinite use lb_default. Defaults to 0.
        ub_default (int, optional): if the upper bound is infinite use ub_default. Defaults to sys.maxsize.

    Returns:
        _type_: _description_
    """
    if interval.lower != P.inf and interval.upper != P.inf:
        lb = interval.lower if interval.left == P.CLOSED else interval.lower + 1
        ub = interval.upper if interval.right == P.CLOSED else interval.upper - 1
        return (lb, ub)
    elif interval.lower != P.inf:
        lb = interval.lower if interval.left == P.CLOSED else interval.lower + 1
        return (lb, ub_default)
    else:
        ub = interval.upper if interval.right == P.CLOSED else interval.upper - 1
        return (lb_default, ub)


def is_singleton(interval):
    """
    Check if interval has only one value

    Args:
        interval (portion): interval

    Returns:
        bool: interval has length 1
    """
    if interval.lower == P.inf or interval.upper == P.inf:
        return False
    if interval.left == P.CLOSED and interval.right == P.CLOSED:
        return interval.lower == interval.upper
    if interval.left == P.OPEN and interval.right == P.CLOSED:
        return interval.lower + 1 == interval.upper
    if interval.left == P.CLOSED and interval.right == P.OPEN:
        return interval.lower == interval.upper - 1
    if interval.left == P.OPEN and interval.right == P.OPEN:
        return interval.lower + 1 == interval.upper - 1


def combine(l, r):
    """
    Given two intervals representing a multiset, combines the two into a single interval
    where elements are indistinguishable only if they were indistinguishable in both

    Args:
        l (IntervalDict): multiset l
        r (IntervalDict): multiset r

    Returns:
        IntervalDict: multiset (l U r)
    """
    # update indistinguishability when and/or domains
    s_dom = l.elements.domain()
    r_int = r.elements.domain()
    if not s_dom.overlaps(r_int):
        return l + r
    else:
        comb = IntervalDict()
        l_keys = l.elements.keys()
        r_keys = r.elements.keys()
        # print(l,"//", r, "   ")
        l_n = 0
        r_n = 0
        while l_n < len(l_keys) or r_n < len(r_keys):
            # iterate partitions on left/right
            # print(l_n, "----", r_n )
            if l_n == len(l_keys):  # add non-overlapping left
                r_int = r_keys[r_n]
                comb = comb.combine(r.elements[r_int], how=is_distinguishable)
                r_n += 1
            elif r_n == len(r_keys):  # add non-overlapping right
                l_int = l_keys[l_n]
                comb = comb.combine(l.elements[l_int], how=is_distinguishable)
                l_n += 1
            else:  # left and right partition share elements
                l_int = l_keys[l_n]
                r_int = r_keys[r_n]
                # print("l:", l_int, "r:", r_int)
                if l_int < r_int:
                    comb = comb.combine(l.elements[l_int], how=is_distinguishable)
                    l_n += 1
                elif r_int < l_int:
                    comb = comb.combine(r.elements[r_int], how=is_distinguishable)
                    comb[r_int] = r.elements[r_int]
                    r_n += 1
                else:
                    # print(l_int,"--",r_int)
                    l_out = l_int - r_int
                    inter = l_int & r_int
                    r_out = r_int - l_int
                    # print("vals", l_out,inter,r_out)
                    if not l_out.empty:
                        if is_singleton(l_out):
                            comb[l_out] = True
                        else:  # partiton of size 1 is always distinguishable
                            comb = comb.combine(
                                l.elements[l_out], how=is_distinguishable
                            )
                    if not r_out.empty:
                        if is_singleton(r_out):
                            comb[r_out] = True
                        else:
                            comb = comb.combine(
                                r.elements[r_out], how=is_distinguishable
                            )
                    if not inter.empty:
                        if is_singleton(inter):
                            comb[inter] = True
                        else:
                            comb = comb.combine(
                                l.elements[inter], how=is_distinguishable
                            )
                            comb = comb.combine(
                                r.elements[inter], how=is_distinguishable
                            )
                    l_n += 1
                    r_n += 1
        return comb


###################
## Logic methods ##
###################


class Not(object):
    """
    Generic Not class, mostly used with DomainFormulas
    """

    def __init__(self, child):
        self.child = child

    def __eq__(self, rhs):
        if isinstance(rhs, Not):
            return self.child == rhs.child
        else:
            return False

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        if isinstance(self.child, str):
            return f"¬{self.child}"
        else:
            return f"¬({self.child})"


class And(object):
    """
    Generic And class with two children
    """

    def __init__(self, l, r):
        self.left = l
        self.right = r

    def __eq__(self, rhs):
        if isinstance(rhs, And):
            same = self.left == rhs.left and self.right == rhs.right
            inverted = self.right == rhs.left and self.left == rhs.right
            return same or inverted
        else:
            return False

    def __hash__(self):
        l = str(self.left)
        r = str(self.right)
        if l[0] < r[0]:
            return hash(l + r)
        else:
            return hash(r + l)

    def __repr__(self):
        # if self.left.name == "":
        #     return str(self.right)
        # elif self.right.name == "":
        #     return str(self.left)
        # else:
        l = ("", "") if isinstance(self.left, str) else ("(", ")")
        r = ("", "") if isinstance(self.right, str) else ("(", ")")
        return f"{l[0]}{self.left}{l[1]} ∧ {r[0]}{self.right}{r[1]}"


class Or(object):
    """
    Generic Or class with two children
    """

    def __init__(self, l, r):
        self.left = l
        self.right = r

    def __hash__(self):
        l = str(self.left)
        r = str(self.right)
        if l[0] < r[0]:
            return hash(l + r)
        else:
            return hash(r + l)

    def __eq__(self, rhs):
        if isinstance(rhs, Or):
            same = self.left == rhs.left and self.right == rhs.right
            inverted = self.right == rhs.left and self.left == rhs.right
            return same or inverted
        else:
            return False

    def __repr__(self):
        l = ("", "") if isinstance(self.left, str) else ("(", ")")
        r = ("", "") if isinstance(self.right, str) else ("(", ")")
        return f"{l[0]}{self.left}{l[1]} ∨ {r[0]}{self.right}{r[1]}"


def dnfy(set):
    """
    Transform a set formula (i.e. And/Or/Not/str) into dnf

    Args:
        set (And/Or/Not/str): a formula describing a set: either a base set or an union/intersection/complement

    Returns:
        And/Or/Not/str: dnf equivalent description of set
    """
    if isinstance(set, str):
        return set
    if isinstance(set, Or):
        ldnfy = dnfy(set.left)
        rdnfy = dnfy(set.right)
        return Or(ldnfy, rdnfy)
    elif isinstance(set, Not):
        if isinstance(set.child, Not):
            return dnfy(set.child.child)
        elif isinstance(set.child, And):
            ldnfy = dnfy(Not(set.child.left))
            rdnfy = dnfy(Not(set.child.right))
            return Or(ldnfy, rdnfy)
        elif isinstance(set.child, Or):
            ldnfy = dnfy(Not(set.child.left))
            rdnfy = dnfy(Not(set.child.right))
            return And(ldnfy, rdnfy)
        else:
            return set
    else:  # isinstance(set, And):
        if isinstance(set.right, Or):
            rdnfy = dnfy(set.right)
            return Or(And(set.right.left, rdnfy), And(set.right.left, rdnfy))
        elif isinstance(set.right, Or):
            ldnfy = dnfy(set.left)
            return Or(And(set.right.left, ldnfy), And(set.right.left, ldnfy))
        else:
            return set


def flatten(op, a):
    """
    Flatten conjunction or disjunction from binary tree to list

    Args:
        op (And/Or): operation type
        a (And/Or/Not/str): root of the tree

    Returns:
        [str/Not/And/Or]: flattened version of a
    """
    if isinstance(a, str) or isinstance(a, Not):
        return [a]
    if isinstance(a, op) and (isinstance(a.left, str) or isinstance(a.left, Not)):
        return [a.left] + flatten(op, a.right)
    elif isinstance(a, op) and (isinstance(a.right, str) or isinstance(a.right, Not)):
        return [a.right] + flatten(op, a.left)
    elif isinstance(a, op) and isinstance(a.left, op) and isinstance(a.right, op):
        return flatten(op, a.left) + flatten(op, a.right)
    else:  # flattening And (Or) but both children are Or (And)
        return a


def nest(op, l):
    """
    Turn a list l into a conjunction or disjunction binary tree
    Args:
        op (And/Or): _description_
        l ([str]): _description_

    Returns:
        op: binary tree where nodes are op and leaves are elements of l
    """
    if len(l) == 1:
        return l[0]
    else:
        return op(l[0], nest(op, l[1:]))


####################
## Labels methods ##
####################


def list2interval(problem, elems, new):
    """
    Convert a list of labels into internal intervals mapped to distinguishable or not
    Args:
        problem (Problem): problem with entities info
        elems ([str]): list of labels of the (multi)set's elements
        dist (Bool): is this a list of distinguishable elements
        new (Bool): if the set is new add elements, otherwise retrieve ids
    Returns:
        IntervalDict: map from intervals to indistinguishability
    """
    intervals = IntervalDict()
    # find duplicates
    if new:
        count = Counter(elems)
        unique = []
        repeated = []
        for label, n in count.items():
            if n > 1:
                indist_elems = [label] + [f"{label}__{i}" for i in range(1, n)]
                repeated.append(indist_elems)
            else:
                unique.append(label)

        # convert entities to intervals
        singletons = entities2intervals(problem, unique, new)
        for iv in singletons:
            intervals[iv] = True
        for indist_set in repeated:
            indist_ivs = entities2intervals(problem, indist_set, new)
            for iv in indist_ivs:
                intervals[iv] = False
    else:  # list of properties without duplicates
        for entity in elems:
            ivs = entities2intervals(problem, [entity], new)
            for iv in ivs:
                dist = (iv.upper - iv.lower) == 0
                intervals[iv] = dist
    return intervals


def entities2intervals(problem, entities, new):
    """
    Convert a list of entities to the internal representation
    Args:
        problem (Problem): problem with entities info
        entities ([str]): labels
        new (Bool): if the set is new add elements, otherwise retrieve ids
    Returns:
        [Int]: The intervals covering the ids corresponding to the labels
    """
    if new:
        lists = [[problem.add_entity(e) for e in entities]]
    else:
        lists = [problem.get_entity(e) for e in entities]
    ints = []
    for list in lists:
        ints += list2Int(list)
    return ints


def list2Int(entities):
    """
    Given a list of entities ids obtains the equivalent intervals
    Args:
        entities ([int]): entities ids

    Returns:
        [Int]: intervals corresponding to the entities' list
    """
    entities.sort()
    ivs = []
    i = 0
    while i < len(entities):
        low = entities[i]
        while i < len(entities) - 1 and entities[i] + 1 == entities[i + 1]:
            i += 1
        hi = entities[i]
        if hi - low >= 1:
            ivs.append(P.closed(low, hi))
        else:
            ivs.append(P.singleton(low))
        i += 1
    return ivs


def simplify_name(elems, complex, universe):
    """
    If a SetFormula resulting from And/Or represents a single element
    then simplify the description using the name of the element

    Args:
        elems (portion): either a single distinguishable element or n indistinguishable copies
        complex (str): the description to use if not one of the simple cases

    Returns:
        str: simplified description
    """
    if is_singleton(elems.domain()):
        if not (True in elems.values()):
            n_elems = sum([1 for _ in P.iterate(elems.domain(), step=1)])
            i = elems.domain().lower
            e_lab = universe.get_label(i, i)
            name = f"{n_elems}x " + str(e_lab)
        else:
            i = elems.domain().lower
            e_lab = universe.get_label(i, i)
            name = f"1x " + str(e_lab)
    else:
        name = complex
    return name



import portion as Int
import math
import itertools

from .configuration import *
from .count import *
from .level_1 import *
from .level_2 import *
from .logger import ProblemLog


class Unsatisfiable(Exception):
    def __init__(self, value, log=None):
        self.value = value
        self.log = log
        if self.log is not None:
            self.log.solution = Zero()

    def __str__(self):
        return repr(self.value)


class SharpCSP(object):
    """
    Counts the possible assignments to a set of vars given choice/count constraints

    Attributes:
    vars : [Variable]
        the variables representing the problem configuration
    config : Configuration
        configuration properties associated to vars
    # agg_f : [AggFormula]
    #     aggrgate constraints on integers
    choice_f : [CPosition]
        choice formulas to enforce
    count_f : [CCounting]
        set of size constraints to enforce
    fixed_choices: int
        subsets only: number of variables already set by choice constraints
    lvl: int
        nesting level for logging
    """

    def __init__(
        self,
        vars,
        config,
        choice_f,
        count_f,
        universe=None,
        lvl=0,
        caption="Subproblem",
        id="1",
        debug=True,
    ):
        # self.agg_f = agg_f
        self.choice_f = choice_f
        self.count_f = count_f
        self.debug = debug
        self.fixed_choices = 0
        self.log = ProblemLog(
            vars,
            config,
            choice_f,
            count_f,
            universe=universe,
            level=lvl,
            id=id,
            caption=caption,
            debug=debug,
        )
        self.id = id
        self.lvl = lvl
        self.n_vars = len(vars)
        self.config = config
        self.vars = vars
        self.universe = universe

    def __str__(self):
        str = "----------\n"
        str += f"Solving {self.n_vars} vars:\n"
        for v in self.vars:
            str += f"\t{v}\n"
        str += "Choice constraints:\n"
        for c in self.choice_f:
            str += f"\t{c}\n"
        str += "Count constraints:\n"
        for c in self.count_f:
            str += f"\t{c}\n"
        # str += "Aggregate constraints:\n"
        # for c in self.agg_f:
        #     str += f"\t{c}\n"
        str += "----------"
        return str

    # def apply_aggs(self):
    #     """
    #     either propagates or counts an aggregate constraint on integers
    #     """
    #     agg = self.agg_f[0]
    #     others = self.agg_f[1:]
    #     if agg.op == min:
    #         return self.propagate_min(agg, others)
    #     else:
    #         return self.propagate_max(agg, others)

    def apply_choice(self, chf):
        """
        Sets variables according to the choice constraints
        """
        self.log.add_relevant_set(chf.formula)
        if isinstance(chf, CPosition):
            if chf.pos - 1 < self.n_vars:
                self.vars[chf.pos - 1] = self.vars[chf.pos - 1] & chf.formula
                # self.debug("Choice set: ", self.vars[chf.pos - 1])
                self.log.propagation(chf)
            else:
                # self.debug(f"Ignoring {chf} because out of range ({self.n_vars})")
                self.log.warning(f"Ignoring {chf} because out of range ({self.n_vars})")

    def apply_count_constraint(self, cc, others):
        """
        Applies the constraints to satisfy 'current' count: for each admissible value n (lb<=n<=ub)
        of the property p in cc, constrain the problem to n variables satisfying p and sum over the
        different (independent) counts

        Args:
            cc (CCounting): counting formula to apply
            others ([CCounting]): remaining counting formulas

        Returns:
            Count: count of the propagation
        """
        # self.debug("Propagating ", cc)
        al = self.log.action(f"Considering constraint {cc}")
        self.log.add_relevant_set(cc.formula)
        out_values = Int.closedopen(0, self.config.size.values.upper) - cc.values
        if cc.values.atomic:
            return self.apply_simple_interval_count(cc, others, al)
        elif out_values.atomic:
            # ignore and then remove the unsat cases
            # self.debug(f"Relax {cc} and remove unsat")
            self.log.detail(al, f"Relax {cc} and remove unsat (={out_values})")
            count_ignore = self.split_on_constraints(1, others, op="add")
            not_cc = CCounting(cc.formula, out_values)
            count_not = self.split_on_constraints(
                1, [not_cc] + others, op="sub", id="2"
            )
            count = count_ignore - count_not
        else:
            # there are multiple separate intervals
            self.log.detail(al, f"Breaking intervals into equalities")
            count = Zero()
            for interval in cc.values:
                partial_cc = CCounting(cc.formula, interval)
                partial_count = self.solve_subproblem(
                    self.vars,
                    self.config,
                    {},
                    [partial_cc] + others,
                    caption=f"Case with {interval} {cc.formula}",
                    op="add",
                    id=str(interval),
                )
                # partial_count = self.apply_count_constraint(partial_cc, others)
                # self.log.add_subproblem("add", partial_count.log)
                count += partial_count
            return count

    def apply_constraints(self):
        """
        If vars are exchangeable we can apply a count otherwise we need to split and consider the combinations of count constraints

        Returns:
            Count: problem sollution
        """
        ex_classes = self.exchangeable_classes()
        if self.config.lvl1():
            if len(ex_classes) > 1:
                count = self.count_non_exchangeable(ex_classes)
            else:
                count = self.count_exchangeable()
        else:
            count = self.shatter_partitions()
        return count

    def apply_inverse_count(self, cc, others, out_values, action):
        """
        If there are more value included than excluded
        ignore and then remove the unsat cases

        Args:
            cc (CCounting): the constraint to be propagated
            others ([Ccounting]): rest of constraints to propagate later
            out_values (Interval): values excluded by the constraint
            action (ActionLog): logging

        Returns:
            Count: nr. solutions with this constraint propagated
        """
        # self.debug(f"Relax {cc} and remove unsat")
        self.log.detail(action, f"Relax {cc} and remove unsat")
        count_ignore = self.split_on_constraints(1, others, op="add")
        not_cc = CCounting(cc.formula, out_values)
        count_not = self.split_on_constraints(1, [not_cc] + others, op="sub", id="2")
        return count_ignore - count_not

    def apply_propositional_count(self, cc, others, ub, action):
        """
        Break interval into equalities with propositionalization of each i in the interval

        Args:
            cc (CCounting): the constraint to be propagated
            others ([Ccounting]): rest of constraints to propagate later
            ub (int): Upper bound from configuration
            action (ActionLog): logging

        Returns:
            Count: nr. solutions with this constraint propagated
        """
        # self.debug(f"Expanding bounds {cc.values}...")
        values = cc.values.replace(upper=ub, right=Int.CLOSED)
        self.log.detail(
            action,
            f"Shatter values {cc.values}: {[i for i in Int.iterate(values, step=1)]}",
        )
        count = Zero()
        # optimization:if we expand values and ...=n is not satisfiable
        # then we can stop beacause ...>= n must be unsat from n on
        stop_at_unsat = False
        for i in Int.iterate(values, step=1):
            if not stop_at_unsat:
                cc_eq = CCounting(cc.formula, Int.singleton(i))
                # self.debug(f"Propagate lower bound #{cc.formula}={lb}...")
                # self.log.detail(al, f"Propagate lower bound #{cc.formula}={lb}...")
                count_case = self.solve_subproblem(
                    self.vars,
                    self.config,
                    {},
                    [cc_eq] + others,
                    caption=f"Case with {i} {cc.formula}",
                    op="add",
                    id=str(i),
                )
                if count_case.is_zero():
                    self.log.detail(action, f"Stopping at {i} which is already 0")
                    stop_at_unsat = True
                count += count_case
        return count

    def apply_simple_count(self, cc, others, action):
        """
        What everything boils down to: propagate the constraint that exactly n
        objects/variables have a property

        Args:
            cc (CCounting): the constraint to be propagated
            others ([Ccounting]): rest of constraints to propagate later
            action (ActionLog): logging stuff

        Returns:
            Count: nr. solutions with this constraint propagated
        """
        try:
            # self.log.detail(al, f"Propagating {cc}")
            choices, prop_vars = self.propagate_cc(cc)
            self.vars = prop_vars
            count = self.split_on_constraints(choices, others)
        except Unsatisfiable as u:
            # self.log.add_subproblem("unsat", self.log)
            # print(self.log)
            count = Zero(tip=u.value)
        return count

    def apply_simple_interval_count(self, cc, others, action):
        """This function collects a few optimizations we can do to count on
            a continuous interval.
        Args:
            cc (CCounting): the constraint with continuous values
            others ([CCounting]): rest of constraints to be propagate
            action (ActionLog): logging stuff

        Returns:
            Count : the count corresponding to propagating this constraint
        """
        max_entities = min(self.n_vars, cc.formula.size())
        # atomic means that there is a single continuous interval
        lb, ub = interval_closed(cc.values, ub_default=max_entities)
        all_values = Int.closed(0, self.n_vars)
        out_values = all_values - cc.values
        if lb == ub:
            return self.apply_simple_count(cc, others, action)
        elif self.config.type != "subset" and (ub >= self.n_vars):
            self.log.detail(
                action,
                f"Fix at least ({cc.values.left}) {cc.values.lower} {cc.formula} and let rest unconstrained",
            )
            return self.apply_simple_count(cc, others, action)
        elif lb == 0 and ub >= self.n_vars:
            # trivial
            return self.split_on_constraints(1, others)
        elif lb == 0 and ub < self.n_vars:
            # propagate n_vars-ub ¬cc.formula and leave rest unconstrained
            ub = self.n_vars - ub
            # self.debug(f"Invert {cc}: at least {ub} {cc.formula.neg()}")
            self.log.detail(action, f"Invert {cc}: at least {ub} {cc.formula.neg()}")
            interval_not = Int.closedopen(ub, Int.inf)
            cc_not = CCounting(cc.formula.neg(), interval_not)
            return self.apply_count_constraint(cc_not, others)
        elif len(out_values) < len(cc.values):
            return self.apply_inverse_count(cc, others, out_values, action)
        else:
            return self.apply_propositional_count(cc, others, ub, action)

    def choose_cc(self):
        """
        Heuristics to decide which counting constraint to propagate when variables are exchangeable

        Returns:
            (CCounting, [CCounting]): the chosen counting constraint to be propagated and the list of the others
        """
        choice = self.count_f[0]
        others = self.count_f[1:]
        ex_class = self.vars[0]
        if isinstance(ex_class, SetFormula):
            for cc in self.count_f:
                same_class = cc.formula == ex_class
                max = cc.values.lower == self.n_vars
                zero = cc.values.upper == 0
                if same_class or max or zero:
                    choice = cc
                    others = self.count_f
                    others.remove(cc)
                    return choice, others
            return choice, others
        else:
            return choice, others

    def compact_ccs(self, counts):
        """
        Simplify some counting constraints, i.e. turn (#p > n), (#p < m) into (#p in [n,m])

        Args:
            counts ([CCounting]): list of counting constraints to be simplified

        Raises:
            Unsatisfiable: if we find some incompatible constraints then counts is unsatisfiable

        Returns:
            [CCounting]: equivalent list of counting constraints compacted
        """
        compact = []
        indexes = range(0, len(counts))
        remove = []
        for i in indexes:
            cc1 = counts[i]
            if cc1.values.empty:
                raise Unsatisfiable("Empty set of valid counts", self.log)
            satisfied, not_satisfied, maybe = self.count_satisfied(cc1)
            if (
                len(maybe) == 0 and len(satisfied) in cc1.values
            ):  # remove ccs already satisfied
                remove += [i]
            else:
                for j in [j for j in indexes if j > i]:
                    cc2 = counts[j]
                    if cc1 != cc2:
                        if cc1.formula == cc2.formula:  # combination is not satisfiable
                            new_int = cc1.values & cc2.values
                            if new_int.empty:
                                raise Unsatisfiable("Conflicting constraints", self.log)
                            else:  # combination satisfiable: merge intervals into single cc
                                merged_cc = CCounting(cc1.formula, new_int)
                                compact.append(merged_cc)
                                remove += [i, j]
                    else:  # cc1 == cc2:
                        # opposite formulas both 0
                        remove += [i]
        keep = [i for i in indexes if not i in remove]
        compact += list(map(lambda i: counts[i], keep))
        final = len(remove) == 0
        result = compact if final else self.compact_ccs(compact)
        return result

    def count_exchangeable(self):
        """
        Take one count constraint and apply it

        Returns:
            Count: the count of solutions of the problem
        """
        # self.debug("Counting exchangeable...")
        cc, others = self.choose_cc()
        count = self.apply_count_constraint(cc, others)
        return count

    def count_non_exchangeable(self, ex_classes):
        """
        If vars are not exchangeable we split the problem into class | rest
        Counting constraints need to be split according to the possible combinations
        Example:
        Split classes french | students with ccs on dutch=2 and french=1
        for each cc prop==n the count can be split between the two classes with i elements on the left and n-i elements on the right, i.e. french==1 | french==0 and french==0 | french==1.
        But with many ccs we need to consider all the combinations of the two so:
        french==0 & dutch==0 | french==1 & dutch==2
        french==1 & dutch==0 | french==0 & dutch==2 ...

        Args:
            ex_classes ([SetFormulas/LiftedSet]): variables grouped by exchangeable classes

        Returns:
            Count: count of the solutions of the problem
        """
        # TODO: optimize by propagating first constraints on sets that are disjoint from others
        split_class_vars, rest_classes_vars = self.split_ex_classes(ex_classes)
        combs_split_class, combs_rest_classes = self.shatter_count_formulas(
            split_class_vars, rest_classes_vars
        )
        al = self.log.action("Shattering constraints")
        tot_count = Zero()
        for i in range(0, len(combs_rest_classes)):
            try:
                comb_split_class = self.compact_ccs(combs_split_class[i])
                comb_rest_classes = self.compact_ccs(combs_rest_classes[i])
                # self.debug(
                #     f"Solving combination {i}: {comb_split_class} // {comb_rest_classes}"
                # )
                self.log.shatter_case(al, i, comb_split_class, comb_rest_classes)
                split_args = [
                    split_class_vars,
                    rest_classes_vars,
                    list(comb_split_class),
                    list(comb_rest_classes),
                ]
                if self.config.lvl1() and not self.config.with_repetition():
                    count = self.split_inj(*split_args)
                elif self.config.lvl1() and self.config.with_repetition():
                    count = self.split(*split_args, shatter_id=i)
                else:
                    count = self.split_partitions(*split_args)
                desc_csc = ", ".join([str(c) for c in comb_split_class])
                desc_crc = ", ".join([str(c) for c in comb_rest_classes])
                self.log.detail(al, f"Split {i+1}: {desc_csc} // {desc_crc}")
                # self.debug(f"Split combination count: {count}")
            except Unsatisfiable as u:
                count = Zero(tip=u.value)
            tot_count += count
        # self.debug(f"Shatter count: {tot_count}")
        return tot_count

    def shatter_count_formulas(self, split_class_vars, rest_classes_vars, ccs=None):
        """
        Returns a shattering of the counting constraints for the given split.
        Given a split, for each size constraints compute the set of splits that shatter the constraint.
        Then combine each different cc shattering with the other constraints' possible splits (cartesian product).
        """
        # self.debug("Shattering size constraints...")
        # self.log.action("Shattering")
        count_f = self.count_f if ccs is None else ccs
        n_split = len(split_class_vars)
        n_rest = self.n_vars - n_split
        ccs_split_class = []
        ccs_rest_classes = []
        for cc in count_f:
            # cases_split_class = []  # split for cc (left)
            # cases_rest_classes = [] # split for cc (right)
            # vals = cc.values
            # if cc.values.upper >= self.n_vars:
            #     interval_any = cc.values.replace(upper=n_split, right=Int.CLOSED)    # adjust upper bound to number of vars
            #     any = CCounting(cc.formula, interval_any)    # add case >= n in split class then work on each =i <n
            #     cases_split_class.append(any)
            #     cc_rest_classes = CCounting(cc.formula, Int.closed(0,n_rest))
            #     cases_rest_classes.append(cc_rest_classes)
            #     if cc.values.left == Int.OPEN:
            #         n_cases = cc.values.lower
            #     else:
            #         n_cases = cc.values.lower-1
            # else:
            #     n_cases = cc.values.upper
            # for i in range(0,n_cases+1):
            # for partition in self.get_feasible_splits(split_class_vars[0],n_split, n_rest, cc.formula):
            #     l_value, r_value = partition
            #     cc_split_class = CCounting(cc.formula, Int.singleton(l_value))
            #     cc_rest_classes = CCounting(cc.formula, Int.singleton(r_value))
            #     # if self.is_feasible_split(split_class_vars[0],n_split,n_rest,cc_split_class,cc_rest_classes):
            #     cases_split_class.append(cc_split_class)
            #     cases_rest_classes.append(cc_rest_classes)
            cases_split_class, cases_rest_classes = self.get_feasible_splits(
                split_class_vars[0], n_split, n_rest, cc
            )
            ccs_split_class.append(cases_split_class)
            ccs_rest_classes.append(cases_rest_classes)
        combs_split_class = list(itertools.product(*ccs_split_class))
        combs_rest_classes = list(itertools.product(*ccs_rest_classes))
        return combs_split_class, combs_rest_classes

    def count(self, vars=None):
        if self.config.lvl1() and self.config.labelled():
            count = self.count_sequence(vars)
        elif self.config.lvl1() and not self.config.labelled():
            count = self.count_subsets(vars)
        elif self.config.lvl2():
            count = self.count_partitions(vars)
        else:
            raise Exception("Unknown configuration type")
        return count

    def count_satisfied(self, cc, var_list=None):
        """Checks for each variable if the property counted

        Args:
            property (SetFormula): _description_
            var_list ([SetFormula]/[LiftedSet], optional): _description_. Defaults to None.
            n (int, optional): how many exchangeable variables need to have the property,
                               useful for checking second level properties. Defaults to 1.

        Returns:
            ([int],[int],[int]): sat/maybe/unsat lists of indexes of variables in each group
        """
        vars = var_list if var_list is not None else self.vars
        property = cc.formula
        sat = []
        not_sat = []
        maybe = []
        if isinstance(property, SetFormula):
            for i, v in enumerate(vars):
                if v in property:
                    sat.append(i)
                elif v.disjoint(property):
                    not_sat.append(i)
                else:
                    maybe.append(i)
        elif isinstance(property, CSize):
            for i, v in enumerate(vars):
                if v.size in property:
                    sat.append(i)
                elif v.size & property == CSize("", Int.empty()):
                    not_sat.append(i)
                else:
                    maybe.append(i)
        else:
            lb, ub = interval_closed(cc.values, ub_default=len(vars))
            for i, v in enumerate(vars):
                satisfies = v.satisfies(property, lb, ub)
                if satisfies is None:
                    maybe.append(i)
                elif satisfies:
                    sat.append(i)
                else:
                    not_sat.append(i)
        return (sat, not_sat, maybe)

    def disjoint(self, classes):
        disj = True
        vars = list(classes.keys())
        indexes = range(0, len(vars))
        for c1 in indexes:
            for c2 in [j for j in indexes if j > c1]:
                disj = disj and vars[c1].disjoint(vars[c2])
        return disj

    def exchangeable_classes(self, var_list=None):
        """
        Groups exchangeable variables in a dict

        Args:
            var_list ([SetFormula/LiftedSet], optional): _description_. Defaults to None.

        Returns:
            {}: _description_
            TODO
        """
        vars = var_list if var_list is not None else self.vars
        classes = {}
        for v in vars:
            new = True
            for c in classes:
                if v == c:
                    classes[c].append(v)
                    new = False
            if new:
                classes[v] = [v]
        return classes

    def filter_domain(self, cases, dformula):
        """
        Given a domain formula dformula and the domain formula case_f of n elements shrinks dformula by (up to) n elements satisfying case_f
        """
        for case in cases:
            n = cases[case]
            inter = dformula & case
            exclude = inter.take(n)
            if exclude.size() > 0:
                df = dformula - exclude
        return df

    def filter_domains(self, cases, variables):
        # self.debug("Filtering domains...")
        al = self.log.action("Filtering domains")
        dfs = variables
        for case in cases:
            # self.debug(f"  Case {case}")
            self.log.detail(al, f"Case {case}")
            n = cases[case]
            if n > 0:
                pool = case
                for df in dfs[1:]:
                    if not df.disjoint(case):
                        pool = pool & df
                exclude = pool.take(n)
                if exclude.size() > 0:
                    updated_dfs = []
                    for df in dfs:
                        # self.debug(f"  Filtering {n} {case} out of {df}:")
                        self.log.detail(al, f"Filtering {n} {case} out of {df}")
                        df -= exclude
                        updated_dfs.append(df)
                    dfs = updated_dfs
                else:
                    dfs = [exclude] * n
        return dfs

    def get_vars(self, indexes):
        return [self.vars[v] for v in indexes]

    def get_subproblem_universe(self, vars, given):
        """Return the correct universe depending on the given one

        Args:
            var (Variable): a variable of the configuration
            given (MultiSet): optional
        """
        ex_classes = self.exchangeable_classes(var_list=vars)
        if given is not None:
            # Promote the multiset to Universe
            u = Universe(
                given.formula,
                given.elements,
                given.name,
                *self.universe.get_labels(),
            )
        elif len(ex_classes) == 1:
            u = Universe(
                vars[0].formula,
                vars[0].elements,
                vars[0].name,
                *self.universe.get_labels(),
            )
        else:
            u = self.universe
        return u

    def histogram(self, var_list=None):
        """
        Returns each exchangeable class with the number of variables belonging to the class
        """
        vars = var_list if var_list is not None else self.vars
        ex_classes = self.exchangeable_classes(vars)
        for c in ex_classes:
            ex_classes[c] = len(ex_classes[c])
        return ex_classes

    def integer_partitions(self, n):
        """
        From http://jeromekelleher.net/generating-integer-partitions.html
        """
        a = [0 for i in range(n + 1)]
        k = 1
        y = n - 1
        while k != 0:
            x = a[k - 1] + 1
            k -= 1
            while 2 * x <= y:
                a[k] = x
                y -= x
                k += 1
            l = k + 1
            while x <= y:
                a[k] = x
                a[l] = y
                yield a[: k + 2]
                x += 1
                y -= 1
            a[k] = x + y
            y = x + y - 1
            yield a[: k + 1]

    def integer_k_partitions(self, n, k):
        """
        Partitions n in k (possibly 0) integers
        """
        parts = []
        for p in self.integer_partitions(n):
            if len(p) < k:
                padded = p + [0] * (k - len(p))
                parts.append(padded)
            elif len(p) == k:
                parts.append(p)
        return parts

    def get_feasible_splits(self, split_class_var, n_split, n_rest, cc):
        disjoint = split_class_var.disjoint(cc.formula)
        # if not isinstance(cc.formula, CCounting):
        if isinstance(cc.formula, SetFormula):
            included = split_class_var in cc.formula
        else:
            # TODO: do the same check for level 2 configurations
            included = False
        cases_split_class = []  # split for cc (left)
        cases_rest_classes = []  # split for cc (right)
        for inter in cc.values:
            lb, ub = interval_closed(inter, ub_default=n_split + n_rest)
            if included and n_split > ub:
                # unsat
                pass
            elif disjoint:
                # rest does all
                if n_rest >= lb:
                    cc_split_class = CCounting(cc.formula, Int.singleton(0))
                    cc_rest_classes = CCounting(cc.formula, Int.closed(lb, n_rest))
                    cases_split_class.append(cc_split_class)
                    cases_rest_classes.append(cc_rest_classes)
                else:
                    # unsat
                    pass
            else:
                if included:
                    lb = max(lb, n_split)  # consider only cases where left>=n_split
                for i in range(lb, ub + 1):
                    partitions = self.integer_k_partitions(i, 2)
                    distributions = partitions + [
                        [r, l] for l, r in partitions if l != r
                    ]
                    for l_val, r_val in distributions:
                        if l_val <= n_split and r_val <= n_rest:
                            cc_split_class = CCounting(cc.formula, Int.singleton(l_val))
                            cc_rest_classes = CCounting(
                                cc.formula, Int.singleton(r_val)
                            )
                            cases_split_class.append(cc_split_class)
                            cases_rest_classes.append(cc_rest_classes)
        return cases_split_class, cases_rest_classes

    def is_feasible_split(self, split_class_var, n_split, n_rest, scc, rcc):
        """
        Checks if we ask to observe more properties than available variables
        Or things that split_class_var cannot satisfy
        """
        self.log.action(
            f"Checking feasibility of splitting {split_class_var} as {scc}//{rcc}"
        )
        if scc.values.empty or rcc.values.empty:
            return False
        slb, sub = interval_closed(scc.values, ub_default=n_split)
        rlb, rub = interval_closed(rcc.values, ub_default=n_rest)
        if slb > n_split:
            # self.debug(f"{scc}//{rcc} is unfeasible because not enough vars in split")
            self.log.detail(
                f"{scc}//{rcc} is unfeasible because not enough vars in split"
            )
            return False
        if rlb > n_rest:
            # self.debug(f"{scc}//{rcc} is unfeasible because not enough vars in rest")
            self.log.detail(
                f"{scc}//{rcc} is unfeasible because not enough vars in rest"
            )
            return False
        disjoint = split_class_var.disjoint(scc.formula)
        included = scc.formula in split_class_var
        if disjoint and slb > 0:
            # self.debug(
            #     f"{scc}//{rcc} is unfeasible because split is disjoint from {scc.formula} and {slb} > 0"
            # )
            self.log.detail(
                f"{scc}//{rcc} is unfeasible because split is disjoint from {scc.formula} and {slb} > 0"
            )
            return False
        if included and sub < n_split:
            # self.debug(
            #     f"{scc}//{rcc} is unfeasible because split is included in {scc.formula} and {sub} < {n_split}"
            # )
            self.log.detail(
                f"{scc}//{rcc} is unfeasible because split is included in {scc.formula} and {sub} < {n_split}"
            )
            return False
        return True

    # def debug(self, s, *args):
    #     if self.debug:
    #         strargs = " ".join(list(map(str, args)))
    #         flat = str(s) + strargs
    #         ind = "\t" * self.lvl
    #         lines = flat.split("\n")
    #         indented = list(map(lambda l: ind + l, lines))
    #         final = "\n".join(indented)
    #         print(final)

    def propagate(self, var, property):
        if isinstance(property, SetFormula):
            return var & property
        elif isinstance(property, CSize):
            return var & LiftedSet(self.universe, property)
        elif isinstance(property, CCounting):
            return var.add_cc(property)
        else:
            raise Exception(
                f"unexpected property type {property}: {type(property)}", self.log
            )

    def propagate_cc(self, cc, var_list=None):
        """
        Given a set of exchangeable variables, propagate one size constraint.
        - Check how many vars already satisfy, maybe satisfy, cannot satisfy the property
        - If there is no var that maybe satisfy, either is already sat or not
        - If there are vars that maybe satisfy, set m vars to satisfy the property where m is how many more entities there need to be, if set the others to not satisfy.
        return the number of exchangeable choices the propagation requires
        """
        al = self.log.propagation(cc)
        vars = var_list if var_list is not None else self.vars
        lb, ub = interval_closed(cc.values, ub_default=len(vars))
        atleast = ub == len(vars)
        satisfied, not_satisfied, maybe = self.count_satisfied(cc, vars)
        diff = lb - len(satisfied)
        var_maybe = [vars[v] for v in maybe]
        if diff == 0 and len(maybe) == 0:
            # self.debug(cc, " already satisfied")
            self.log.detail(al, f"is already satisfied")
            return 1, vars
        elif len(maybe) == 0:
            # self.debug(cc, " is unsat here")
            self.log.detail(al, "is unsat with these variables")
            raise Unsatisfiable(
                f"Stopped while trying to propagate {cc} (conflicting with other constraints)"
            )
        else:
            v = var_maybe[0]
            # self.debug(f"{len(maybe)} exchangeable constrainable vars: {v}")
            self.log.detail(al, f"{len(maybe)} exchangeable constrainable vars: {v}")
            if not self.config.labelled():
                n_choices = 1
            else:
                n_choices = math.comb(len(maybe), abs(diff))
            sat_f = self.propagate(v, cc.formula)
            not_sat_f = self.propagate(v, cc.formula.neg())
            for i in maybe:
                if diff < 0 and not atleast:
                    vars[i] = not_sat_f
                    diff += 1
                elif diff > 0:
                    vars[i] = sat_f
                    diff -= 1
                else:  # diff=0
                    if not atleast:
                        vars[i] = not_sat_f
            if diff != 0 and not atleast:
                self.log.detail(al, "is unsat with these variables")
                raise Unsatisfiable(f"Cannot satisfy {cc} :(")
            else:
                self.log.detail(al, self.vars)
                return n_choices, vars

    def relevant_cases_intersection(self, universe, rest_classes):
        """
        Computes recursively the intersections of relevant cases/domains
        """
        first = rest_classes[0]
        if first == universe:
            return self.relevant_cases_intersection(universe, rest_classes[1:])
        combinations = [[first, first.neg()]]
        for dom in rest_classes[1:]:
            # if we have other subsets we can ignore the universe since there is no element
            # in not(universe)
            if not dom == universe:
                comb = [dom, dom.neg()]
                combinations.append(comb)
        combinations = list(itertools.product(*combinations))
        cases = []
        for c in combinations:
            dom_base = c[0]
            if len(c) > 1:
                for dom in c[1:]:
                    dom_base = dom_base & dom
            if dom_base.size() > 0:
                cases.append(dom_base)
        return cases

    def relevant_cases(self, split_df, rest_classes):
        """
        Computes the cases for elements of type split_df w.rt. rest_classes. i.e. if we split on french and we have other classes like dutch and italian,
        we need to consider for alldiff the cases where some of the french are both/neither dutch or italian or one of the two
        """
        # self.debug("Computing case combinations of relevant classes:")
        al = self.log.action("Computing case combinations of relevant classes")
        # if we have one other class consider only positive, since n out of m
        # positives is the same as m-n negatives out of m
        if len(rest_classes) == 0:
            res = []
        elif len(rest_classes) == 1:
            first = rest_classes[0]
            if first != self.universe:
                res = [first, first.neg()]
            else:
                res = [split_df]
        else:
            res = self.relevant_cases_intersection(split_df.universe, rest_classes)
        self.log.detail(al, res)
        for r in res:
            self.log.add_relevant_set(r)
        # self.debug(f"\t{res}")
        return res

    def solve(self):
        for c in self.choice_f:
            self.apply_choice(c)
        try:
            self.count_f = self.compact_ccs(self.count_f)
        except Unsatisfiable as u:
            return Zero(tip=u.value)
        if len(self.count_f) > 0:
            count = self.apply_constraints()
        # elif len(self.agg_f) > 0:
        #     count = self.apply_aggs()
        else:
            count = self.count()
        # self.debug("=========")
        # self.debug("Solution:" + str(count))
        return Solution(count, self.log)

    def split(
        self,
        split_class_vars,
        rest_classes_vars,
        split_class_ccs,
        rest_classes_ccs,
        shatter_id=1,
    ):
        scv = [v.copy() for v in split_class_vars]
        rcv = [v.copy() for v in rest_classes_vars]
        # self.debug(f"Split class :")
        al = self.log.action("Splitting")
        split_class_count = self.solve_subproblem(
            scv,
            self.config,
            [],
            split_class_ccs,
            op="split-left",
            # id=f"{id}",
            shatter_id=shatter_id,
            caption="Split class",
        )
        if split_class_count.is_zero():
            self.log.detail("Left split is 0: ignoring right")
            return split_class_count
        else:
            # self.debug("Rest class: ")
            rest_classes_count = self.solve_subproblem(
                rcv,
                self.config,
                [],
                rest_classes_ccs,
                op="split-right",
                # id=f"{id}",
                shatter_id=shatter_id,
                caption="Rest class",
            )
            count = split_class_count * rest_classes_count
            # self.debug("==========")
            return count

    def solve_subproblem(
        self,
        vars,
        config,
        choice_constr,
        count_constr,
        universe=None,
        op="none",
        caption="Subproblem",
        id="1",
        shatter_id=None,
    ):
        """
        Recursive call to the solver

        Args:
            vars ([Variable]): variables initialization
            config (str): description of the configuration config
            choice_constr ([PConstraint]): positional constraints
            count_constr ([CConstraint]): counting constraints

        Returns:
            Count: solution of the subproblem
        """
        # self.debug(f"\tSubproblem ({config}):")
        vars = [v.copy() for v in vars]

        sub_id = f"{self.id}.{id}"
        if op.startswith("split"):
            if op.endswith("left"):
                sub_id += "L"
            else:
                sub_id += "R"

        subproblem = SharpCSP(
            vars,
            config,
            choice_constr,
            count_constr,
            self.get_subproblem_universe(vars, universe),
            self.lvl + 1,
            caption,
            id=sub_id,
            debug=self.debug,
        )
        try:
            sub_sol = subproblem.solve()

            # logging
            if shatter_id is None:
                self.log.add_subproblem(op, sub_sol.log)
            else:
                if op == "split-left":
                    self.log.add_split_left(sub_sol.log, shatter_id)
                else:
                    self.log.add_split_right(sub_sol.log, shatter_id)
            return sub_sol.count
        except Unsatisfiable as u:
            # self.debug("\t==========\n\tUnsat: 0")
            return Zero(tip=u.value)

    def split_on_constraints(self, n_choices, others, id="1"):
        # self.debug("Splitting on other constraints...")
        al = self.log.action("Splitting on other constraints")
        if len(others) == 0:
            # self.debug("... no other constraints")
            self.log.detail(al, "no other constraints")
            count = self.count()
        else:
            count = self.solve_subproblem(
                self.vars,
                self.config,
                [],
                others,
                op="none",
                caption=self.log.caption,
                id=id,
            )
        count.log = self.log
        res = count.with_choices(n_choices)
        return res

    def split_ex_classes(self, ex_classes):
        """
        Groups the exchangeable classes as left | right by taking one as left and uniting the others
        don't pick universe as split class
        """
        split_class = None
        rest_classes = []
        if self.config.lvl2():
            it_excls = iter(ex_classes)
            split_class = next(it_excls)
            for i in it_excls:
                rest_classes = rest_classes + ex_classes[i]
        else:
            for ex_class in ex_classes:
                if split_class is None and ex_class != self.universe:
                    split_class = ex_class
                else:
                    rest_classes = rest_classes + ex_classes[ex_class]
        # it_excls = iter(ex_classes)
        # split_class = next(it_excls)
        # rest_classes = []
        # for i in it_excls:
        #     rest_classes = rest_classes + ex_classes[i]
        return (ex_classes[split_class], rest_classes)

    #######################
    ## Level 1 methods ##
    #######################

    def count_multisubsets_exchangeable(self, var_list=None):
        """
        Select with replacement from a multiset M: that is equivalent to select with repetition from the support of M:
        find number of entities and apply binomial coefficient

        Args:
            var_list ([SetFormulas]], optional): List of variables representing the mset. Defaults to None.

        Returns:
            Count: count
        """
        vars = var_list if var_list is not None else self.vars
        n = len(vars)
        dom = vars[0]
        indist = dom.elements.find(False)
        indist_sizes = self.get_sizes_indistinguishable(indist)
        indist_size = sum(indist_sizes)
        dist_size = dom.size() - indist_size + len(indist_sizes)
        m = dist_size + n - 1
        # self.debug(f"Multisubsets with all exchangeable: ({dist_size+n-1},{n})={count}")
        al = self.log.action(f"Counting multisubsets with all exchangeable")
        self.log.detail(al, f"With binomial coefficient ({m} {n})")
        sol = Binomial(n, m, f"Choose {n} objects out of {m} distinguishable objects")
        return sol

    def count_multisubsets_non_exchangeable(self, var_list=None):
        """If there are non-exchangeable variables in order to exploit exchangeability classes have to be disjoint.
        Otherwise multiplication is introducing some order. So if dijoint apply multiplication otherwise we need to
        refine the variables that are not disjoint with the relevant parts.

        Args:
            var_list ([SetFormula], optional): List of variables representing the mset. Defaults to None.

        Returns:
            Count: count
        """
        # self.debug("Multisubsets with non-exchangeable...")
        al = self.log.action("Multisubsets with non-exchangeable...")
        vars = var_list if var_list is not None else self.vars
        ex_classes = self.exchangeable_classes(vars)
        if self.disjoint(ex_classes):
            # self.debug("... but disjoint")
            self.log.detail(al, "... but disjoint")
            count = One()
            for exc in ex_classes:
                s = self.count_multisubsets_exchangeable(ex_classes[exc])
                count *= s
                self.log.detail(al, "Product of multisets of each class")
        else:
            # if classes are not disjoint no multiplication rule
            # self.debug("... and not disjoint: disjoining with relevant parts")
            self.log.detail(al, "... and not disjoint: disjoining with relevant parts")
            relevant = self.relevant_cases_intersection(
                self.universe, list(ex_classes.keys())
            )
            unsplit = [ec for ec in ex_classes if ec not in relevant]
            decompose_class = unsplit[0]
            count = self.shannon_indep_multisubsets(decompose_class, relevant, 0)
        return count

    def shannon_indep_multisubsets(
        self, decompose_class, relevant, previous, var_list=None
    ):
        """If we have two non-disjoit classes for example universe U and a domain D and variables D U U U.
        Then multiplication cannot be applied and we need to refine U. So for each variable that is not a relevant
        part (D or ¬D) we specialize it with the options, e.g. D D U U and D ¬D U U. We do it recursively but
        we are introducing an order, e.g. we can get D D D ¬D or D ¬D D D or D D ¬D D and many other exchangeable combinations.
        With previous we keep track of the last id of relevant part used and we make sure we use only later ones for new
        specializations.

        Args:
            decompose_class (SetFormula): the domain/exchangeability class to be specialized
            relevant ([DomainFOrmula]): the list of partitions that we use to specialize the vars in dijoint domains
            previous (int): index of the last
            var_list ([SetFormula], optional): List of variables representing the mset. Defaults to None.

        Returns:
            Count: count
        """
        vars = var_list if var_list is not None else self.vars
        ex_classes = self.exchangeable_classes(vars)
        self.log.action("Shattering multiset choices")
        count = Zero()
        for i, r in enumerate(
            relevant
        ):  # specialize variable with each different relevant set
            inter = decompose_class & r  #
            n = len(ex_classes[decompose_class])
            if (
                inter.size() > 0 and i >= previous
            ):  # if the specialized variable is empty or we are not following the order skip
                vars = [
                    v for k in ex_classes for v in ex_classes[k] if k != decompose_class
                ]
                others = [decompose_class] * (
                    n - 1
                )  # specialize one variable with this r then the others recursively if necessary
                vars = vars + [inter] + others
                new_ex_classes = self.exchangeable_classes(vars)
                if len(new_ex_classes) == 1:
                    count += self.count_multisubsets_exchangeable(vars)
                elif self.disjoint(new_ex_classes):
                    count += self.count_multisubsets_non_exchangeable(vars)
                else:
                    unsplit = [ec for ec in new_ex_classes if ec not in relevant]
                    next_decompose_class = unsplit[0]
                    count += self.shannon_indep_multisubsets(
                        next_decompose_class, relevant, i, var_list=vars
                    )
        return count

    def count_sequence(self, var_list=None):
        """
        Look at a sequence problem and figure out what to do:
        - Sequence: cartesian product
        - Sequence with alldiff (permutation):
            - Exchangeable: counting rule (factorial)
            - Non-exchangeable: split/shatter
        """
        vars = var_list if var_list is not None else self.vars
        if self.config.type == "sequence":
            # self.debug("Counting sequences:")
            sol = self.count_sequence_any(vars)
        else:
            # self.debug("Counting permutations:")
            al = self.log.action("Counting permutations")
            ex_classes = self.exchangeable_classes(vars)
            disjoint_classes = self.disjoint(ex_classes)
            if len(ex_classes) == 1:
                # self.debug(f"Exchangeable variables...")
                self.log.detail(al, f"Exchangeable variables")
                sol = self.count_permutation_exchangeable(vars)
            # elif disjoint_classes:
            #     self.log("Different classes but disjoint...")
            #     sol = self.count_sequence_cartesian(vars)
            else:
                # self.debug("Splitting injectivity...")
                self.log.detail(al, "Splitting injectivity")
                scv, rcv = self.split_ex_classes(ex_classes)
                sol = self.split_inj(scv, rcv, [], [])
        return sol

    def count_sequence_any(self, var_list=None):
        """
        When variables are independent, solutions are from the cartesian product of the domain.
        Indistinguishable elements collapse to one in terms of possible choiches
        """
        vars = var_list if var_list is not None else self.vars
        count = One()
        al = self.log.action("Counting regular sequence")
        for v in vars:
            # the available choices are all dist. elems + each distinguishable property
            indist = v.elements.find(False)
            indist_sizes = self.get_sizes_indistinguishable(indist)
            indist_size = len([s for s in indist_sizes if s > 0])
            dist_size = v.size() - sum(
                indist_sizes
            )  # do not count each indistinguishable as a different choice
            # self.debug(f"\t{dist_size} different choices for {v}")
            self.log.detail(al, f"\t{dist_size} different choices for {v}")
            # but each set of indistinguishable counts as one possible choice
            choices = dist_size + indist_size
            count *= Count(choices, subproblems=0)
        # self.debug(f"\tDomain product: {count}")
        self.log.detail(al, f"\tDomain product: {count}")
        return count

    def count_permutation_exchangeable(self, var_list=None):
        """
        When all elements belong to the same set in a permutation, apply falling factorial
        However if sets have indistinguishable elements, account for it
        """
        vars = var_list if var_list is not None else self.vars
        n = len(vars)
        dom = vars[0]
        size = dom.size()
        extra = size - n
        al = self.log.action("Regular permutation")
        if extra >= 0:  # are there enough elements to fill the permutation?
            indist = dom.elements.find(False)
            dist = dom.elements.find(True)
            dist_lab = self.universe.get_label(dist)
            if indist.empty:
                self.log.detail(
                    al, f"Without indistinguishable: applying falling factorial"
                )
                if extra == 0:
                    count = Factorial(size, "Nr. orders for all objects")
                else:
                    num = Factorial(size, "Nr. orders for all objects")
                    denom = Factorial(extra, "Nr. excluded objects")
                    count = Divide(num, denom)
                count.set_histogram(self.histogram())
            else:
                self.log.detail(
                    al,
                    f"With indistinguishable elements: considering distributions of indistinguishable",
                )
                indist_sizes = self.get_sizes_indistinguishable(indist)
                dist_size = size - sum(indist_sizes)
                choices = self.get_group_choices([dist_size] + indist_sizes, n)
                labels = [self.universe.get_label(i) for i in indist]
                count = Zero()
                for choice in choices:
                    # each combination of choices is a valid disjoint subproblem
                    dist_choice = choice[0]
                    n_dist_choices = Binomial(
                        dist_size,
                        dist_choice,
                        f"Choose {dist_choice} of {dist_size} (distinguishable) {dist_lab} for {n} object(s)",
                    )
                    arrange_all = Factorial(
                        n,
                        f"Permutations of {n} {dom.name}",
                    )  # count = n! / n1!*n2!*...*nm!
                    arrange_indist_choices = [
                        Factorial(
                            ic,
                            f"Extra permutations of (indist.) {labels[i]}",
                        )
                        for i, ic in enumerate(choice[1:])
                    ]
                    prod_indist_choices = One()
                    for aic in arrange_indist_choices:
                        prod_indist_choices *= aic
                    count += Divide(
                        n_dist_choices * arrange_all,
                        prod_indist_choices,
                        histogram=self.histogram(),
                    )

        else:  # not enough elements in dom to make a n_vars long permutation
            count = Zero()
            self.log.detail(al, f"{len(vars)} different vars for {size} values!")
        return count

    def count_subsets(self, var_list=None):
        """
        Look at a subset problem and figure out what to do:
        - Subset with repetition: counting rule
        - Subset with alldiff (no repetition):
            - Exchangeable: counting rule (binomial coefficient)
            - Non-exchangeable: split/shatter
        """
        # self.log.action("Counting (multi)subsets")
        # self.log.detail(al,f"There are {self.fixed_choices} fixed elements")
        vars = var_list if var_list is not None else self.vars
        f = self.fixed_choices
        if self.config.type == "multisubset":
            ex_classes = self.exchangeable_classes(vars)
            if len(ex_classes) == 1:
                sol = self.count_multisubsets_exchangeable(vars[f:])
            else:
                sol = self.count_multisubsets_non_exchangeable(vars[f:])
        else:
            ex_classes = self.exchangeable_classes(vars)
            if len(ex_classes) == 1:
                sol = self.count_subsets_exchangeable(vars[f:])
            else:
                # self.debug("Splitting injectivity...")
                scv, rcv = self.split_ex_classes(ex_classes)
                sol = self.split_inj(scv, rcv, [], [])
        return sol

    def count_subsets_exchangeable(self, var_list=None):
        """
        When all elements belong to the same set in a permutation, apply binomial coefficient
        However if sets have indistinguishable elements, account for it
        """
        vars = var_list if var_list is not None else self.vars
        n = len(vars)
        dom = vars[0]
        size = dom.size()
        extra = size - n
        al = self.log.action("Counting regular subsets")
        if extra >= 0:
            indist = dom.elements.find(False)
            if indist.empty:
                count = Binomial(size, n, f"Choose {n} from {size} {dom.name} objects")
                self.log.detail(al, f"With binomial coefficient: ({size} {n})")
            else:
                indist_sizes = self.get_sizes_indistinguishable(indist)
                dist_size = size - sum(indist_sizes)
                choices = self.get_group_choices([dist_size] + indist_sizes, n)
                count = Zero()
                self.log.detail(
                    al, f"With indistinguishable elements: accounting for histograms"
                )
                self.log.detail(al, f"Indistinguishable groups: {indist}")
                for choice in choices:
                    self.log.detail(al, str(choice))
                    dist_choice = choice[0]
                    n_dist_choices = Binomial(
                        dist_size,
                        dist_choice,
                        f"Choose {dist_choice} from {dist_size} {dom.name} objects",
                    )
                    count += n_dist_choices
                # self.debug(f"Subsets with indistinguishable elements: {count}")
        else:  # not enough elements in dom to make a n sized subset
            self.log.detail(al, f"{len(vars)} different vars for {size} values!")
            count = Zero()
            # self.debug(f"{len(vars)} different vars for {size} values!")
        count.set_histogram(self.histogram())
        return count

    def get_sizes_indistinguishable(self, indistinguishable):
        """
        Given an interval containing the indistinguishable elements, returns
        the count of each group
        """
        indist_sizes = []
        for left, lower, upper, right in Int.to_data(indistinguishable):
            if not left:
                lower += 1
            if not right:
                upper -= 1
            if upper > lower:
                size = upper - lower + 1
            else:
                size = 0
            indist_sizes.append(size)
        return indist_sizes

    def get_group_choices(self, sizes, length):
        """
        For a sequence of given length, fix any way of picking n elements from each distinguishable property,
        a distinguishable property is either a distinguishable element (1 choice from it) or a set of
        indistinguishable elements (#set possible choices)
        sizes: size of each group of elements in the domain:
               first one is the group of distinguishable, others are groups of indistinguishable
        length: length of the sequence
        """
        if len(sizes) > 0:
            group_size = sizes[0]
            upper = min(length, group_size)
            lower = max(0, length - sum(sizes[1:]))
            indist_choices = []
            for s in range(
                lower, upper + 1
            ):  # for each set we can pick from 0 to #set elements
                rest_indist_choices = self.get_group_choices(sizes[1:], length - s)
                indist_choices += [[s] + ric for ric in rest_indist_choices]
            return indist_choices
        else:
            return [[]]

    def split_inj(
        self, split_class_vars, rest_classes_vars, split_class_ccs, rest_classes_ccs
    ):
        """
        Splits the problem under injectivity: we split between one split class and rest: split class has n variables and rest has some other classes/properties: we need to count how many "relevant" properties can end up in the split class. The relevant properties are the formulas that we have in rest: if we split between french and dutch, when splitting on french we consider the cases where n french speakers were dutch speakers, and adjust the domain of dutch in the "rest" problem.
        """
        split_class = split_class_vars[0]
        # self.debug(f"Split class: {split_class}")
        al = self.log.action(f"Splitting injectivity on class {split_class}")
        rest_domains = self.exchangeable_classes(rest_classes_vars).keys()
        relevant_rest_domains = [
            d for d in rest_domains if not split_class.disjoint(d)
        ]  # optimization: filter out disjoint domains
        indist_domains = [
            SetFormula(
                split_class.formula,
                split_class.elements[dom],
                split_class.universe,
                split_class.name,
            )
            for dom in split_class.elements.find(False)
        ]
        cases = self.relevant_cases(split_class, indist_domains + relevant_rest_domains)
        c = len(cases)
        if c == 0:
            # optimization: if all disjoint then split class subproblem already independent
            # self.debug(
            #     "Split class is disjoint from others: no need for injectivity split..."
            # )
            self.log.detail(
                al,
                "Split class is disjoint from others: no need for injectivity split",
            )
            count = self.split(
                split_class_vars, rest_classes_vars, split_class_ccs, rest_classes_ccs
            )
        else:
            n = len(split_class_vars)
            possible_partitions = self.integer_k_partitions(n, c)
            count = Zero()
            for partition in possible_partitions:
                possible_composition = set(itertools.permutations(partition))
                for id, composition in enumerate(possible_composition):
                    count += self.split_inj_case(
                        id,
                        composition,
                        cases,
                        split_class_vars,
                        rest_classes_vars,
                        split_class_ccs,
                        rest_classes_ccs,
                    )
        return count

    def split_inj_case(
        self,
        id,
        composition,
        cases,
        split_class_vars,
        rest_classes_vars,
        split_class_ccs,
        rest_classes_ccs,
    ):
        # self.debug(f"Case {n_case} of {cases}")
        split_count_formulas = split_class_ccs.copy()
        split_inj_formulas = [
            CCounting(cases[i], Int.singleton(composition[i]))
            for i in range(0, len(composition))
        ]
        split_count_formulas = split_count_formulas + split_inj_formulas
        # self.debug(
        #     f"Case {n} {split_class} are s.t. {split_count_formulas}:"
        # )
        try:
            # log info
            n = len(split_class_vars)
            v = split_class_vars[0]
            split_count_formulas = self.compact_ccs(split_count_formulas)
            description = f"Left split: case {n} {v} are s.t. {split_count_formulas}"

            split_class_count = self.solve_subproblem(
                split_class_vars,
                self.config,
                [],
                split_count_formulas,
                op="split-left",
                caption=description,
                id=f"{id+1}",
                shatter_id=id + 1,
            )
            if split_class_count.is_zero():
                # optimization: do not solve rest if split already unsat
                return Zero()
            for j, count_hist in enumerate(split_class_count.histograms):
                # consider all possible histograms for a left solution
                # i.e. compositions over relevant sets in a valid left configuration

                # update universe with left problem's histogram info
                c, hst = count_hist
                filtered = self.filter_domains(hst, rest_classes_vars)
                new_univ = self.filter_domains(hst, [self.universe])[0]

                # log info
                desc_hist = ", ".join([f"{n} {set}" for set, n in hst.items()])
                description = f"Right split removing {desc_hist}"

                # multiplication rule
                left_count = Count(c, histogram=hst)
                right_count = self.solve_subproblem(
                    filtered,
                    self.config,
                    [],
                    rest_classes_ccs,
                    universe=new_univ,
                    op="split-right",
                    caption=description,
                    id=f"{id+1}.H{j+1}",
                    shatter_id=id + 1,
                )
                # self.debug(
                #     f"Split result = {c} * {rest_count.count} = {partial_count * rest_count}"
                # )
                return left_count * right_count
        except Unsatisfiable as u:
            return Zero(tip=u.value)
            # self.debug(f"\t is unsat.")

    #########################
    ##  Level 2 methods ##
    #########################

    def count_exchangeable_class_partitions(
        self, size, n_elems, n_partitions, formula="universe"
    ):
        # https://math.stackexchange.com/questions/640558/how-many-ways-can-n-elements-be-partitioned-into-subsets-of-size-k
        self.log.action(
            f"Dividing {n_elems} from {formula} in {n_partitions} {self.config.type} of size {size}"
        )
        if size == 0:
            return One()
        pick = size * n_partitions
        count_pick = Factorial(
            pick,
            f"Nr. of ways to write a sequence of the {pick} picked objects where each block of {size} objects is a part",
        )
        # count_pick = Solution(
        #     math.factorial(pick),
        #     [],
        #     self.log,
        #     0,
        #     f"\\texttip{{ {pick}! }}{{Nr. of ways to write a sequence of the {pick} picked objects where each block of {size} objects is a part}}",
        # )
        count_pick_choice = Binomial(
            n_elems, pick, f"Pick {pick} of {formula} out of {n_elems}"
        )
        # pick_choice = math.comb(n_elems, pick)
        # count_pick_choice = Solution(
        #     pick_choice,
        #     [],
        #     self.log,
        #     0,
        #     f"\\binom{{ \\texttip{{ {n_elems} }}{{ Total nr. of objects}} }} {{ \\texttip{{ {pick} }}{{Nr. of objects to distribute }} }}",
        # )
        # counting rule
        # d1 = math.factorial(size) ** n_partitions
        # d2 = math.factorial(n_partitions)
        # count_d1 = Solution(
        #     d1,
        #     [],
        #     self.log,
        #     0,
        #     f"\texttip{{ ({size}!)^{n_partitions} }}{{ Each of the {n_partitions} has {size} internal orderings}}",
        # )
        # count_d2 = Solution(
        #     d2,
        #     [],
        #     self.log,
        #     0,
        #     f"\\texttip{{ {n_partitions}! }}{{ Nr. of external orderings of the {n_partitions} parts }}",
        # )
        d1 = Factorial(size, f"Internal ordering of each partition") ** Count(
            n_partitions, tip="Number of parts"
        )
        d2 = Factorial(
            n_partitions,
            tip=f"Nr. of external orderings of the {n_partitions} parts",
        )
        if self.config.type == "composition":
            count = Divide(
                count_pick, d1, tip="Total permutations divided by nr. internal orders"
            )
        else:
            count = Divide(
                count_pick,
                (d1 * d2),
                tip="Total permutations divided by nr. internal and external orders",
            )
        # TODO: with indist elements?
        # if self.type == "composition":
        #     count = count_pick // count_d1
        # else:
        #     count = count_pick // (count_d1 * count_d2)
        return count_pick_choice * count

    def count_fixed_partitions(
        self, relevant, var_list=None, caption="", id="1", log=None
    ):
        """
        If everything is fixed, we can count subsets like fixed sizes, i.e.
        for each partition pick the given number of p stated by (#p = n) and remove them from universe
        Account for exchangeability of variables in the process: pick at exchangeable class level
        """
        vars = var_list if var_list is not None else self.vars
        log = self.log if log is None else log
        al = log.action("Counting fixed partitions from histograms")
        for k in vars[0].histogram.keys():
            log.detail(al, str(k))
            log.add_relevant_set(k)
        choices = {rvs: rvs.size() for rvs in relevant}
        count = One()
        for i, v in enumerate(vars):
            al2 = log.action(
                f"Histogram for set {i+1} = {[val for val in v.histogram.values()]}"
            )
            for rvs in v.histogram:
                if not rvs.all_indistinguishable():
                    n = choices.get(rvs, 1)
                    k = v.histogram[rvs]
                    ec_choices = Binomial(n, k, tip=f"Choose {k} out of {n} {rvs}")
                    if rvs in choices:
                        choices[rvs] = n - k
                    count *= ec_choices
            log.detail(al2, f"Count = {count}")
        if self.config.type == "partition":
            part_histograms = [frozenset(v.histogram.items()) for v in vars]
            ex_hists = self.histogram(part_histograms)
            for hc in ex_hists:  # account for overcounting exchangeable choices (ugly)
                n_exchangeable = Factorial(ex_hists[hc], tip="Nr. exchangeable parts")
                count = Divide(count, n_exchangeable)
            log.detail(al, f"Count = {count}")
        return count

    def count_partitions(self, var_list=None, id=0):
        """
        Look at a partition problem and figure out what to do:
        - If there are no constraints and everything is exchangeable: counting rule (Stirling)
        - If there are size constraints solve with simpler solving method
        - If there are size constraints use the integer constraint solver method
        - If there are indistinguishable elements use the integer constraint solver method
        """
        vars = var_list if var_list is not None else self.vars
        p_log = ProblemLog(
            vars,
            universe=self.universe,
            caption=f"Constraint combination nr. {id}",
            id=id,
            level=self.lvl + 1,
            config=self.config,
            debug=self.debug,
        )
        n_elems = self.universe.size()
        ex_classes = self.exchangeable_classes(vars)
        counting_constraints = False
        for v in vars:
            counting_constraints = counting_constraints or len(v.ccs) > 0
        no_indist = self.universe.elements.find(False).empty
        if len(ex_classes) == 1:
            partition = vars[0]
            max_size = self.universe.size() - self.n_vars + 1
            unconstrained = CSize("any", Int.closed(1, max_size))
            if (
                not counting_constraints
                and partition.size == unconstrained
                and no_indist
            ):
                # self.debug(
                #     "Exchangeable variables and no constraints: Stirling numbers 2nd kind"
                # )
                al = p_log.action(
                    "Exchangeable variables and no constraints: Stirling numbers 2nd kind"
                )
                count = Stirling(
                    n_elems,
                    len(vars),
                    tip=f"Stirling number of the second kind: nr. partitions of {n_elems} distinguishable objecsts in {len(vars)} non-empty groups",
                )
                if self.config.type == "composition":
                    perm = Factorial(
                        len(vars), f"Nr. permutations of {len(vars)} groups"
                    )
                    count *= perm
            else:
                count = self.count_constrained_partitions(vars, log=p_log)
        else:
            count = self.count_constrained_partitions(vars, log=p_log)
        # self.debug(f"Count: {count}")
        return Solution(count, p_log)

    def count_constrained_partitions(self, var_list=None, log=None):
        """
        If all relevant quantities are fixed then count.
        Otherwise decompose intervals in constraints into equalities (accounting for all possible combinations)
        """
        # self.debug("There are constraints: consider relevant sets")
        log = self.log if log is None else log
        al = log.action("There are constraints: consider relevant sets")
        vars = var_list if var_list is not None else self.vars
        cases = self.universe.indistinguishable_subsets()
        for v in vars:
            cases = cases.union(v.relevant())
        cases = list(cases)
        if len(cases) == 1 and cases[0] == self.universe:
            relevant = cases
        else:
            relevant = self.relevant_cases_intersection(self.universe, cases)
        # self.debug(f"Relevant parts: {relevant}")
        log.detail(al, relevant)
        if self.fixed_vars(vars, relevant):
            return self.count_fixed_partitions(relevant, vars, log)
        else:
            return self.count_unfixed_partitions(relevant, vars, log)

    def fixed_vars(self, vars, relevant):
        """Check if everything relevant is fixed in each part/variable

        Args:
            vars ([LiftedSet]): parts
            relevant ([MultiSet]): list of relevant sets

        Return:
            Bool : all fixed
        """
        fixed = True
        for v in vars:
            fixed_size = is_singleton(v.size.values)
            fixed = fixed and fixed_size
            v.histogram = {rv_set: -1 for rv_set in relevant}
            if len(v.histogram) == 1 and fixed_size:  # register fix from size
                v.histogram[self.universe] = v.size.values.lower
            else:  # register fix from constraints
                for rv_set in relevant:
                    fixed_rv_set = False
                    for cc in v.ccs:
                        single = is_singleton(cc.values)
                        if rv_set == cc.formula and single:
                            v.histogram[rv_set] = cc.values.lower
                            fixed_rv_set = True
                    fixed = fixed and fixed_rv_set
        return fixed

    def fix_exchangeable_partititons(self, rv_set, n, prev_e=1, var_list=None):
        vars = var_list if var_list is not None else self.vars
        k = len(vars)
        # self.debug(f"Fixing {n} of {rv_set} on {k} exchangeable")
        # self.debug(f" {vars[0]}")
        al = self.log.action(f"Fixing {n} of {rv_set} on {k} exchangeable sets")
        self.log.detail(al, f"Exchangeable class: {vars[0]}")
        ips = self.feasible_relevant_partitions(rv_set, n, vars)
        valid = []
        for int_partition in ips:
            prop_vars = []
            for i, v in enumerate(vars):
                new_v = v.copy()
                new_v.histogram[rv_set] = int_partition[i]
                prop_vars.append(new_v)
            if len(prop_vars) == k:
                e = (
                    self.n_int_perms(int_partition)
                    if self.config.type == "composition"
                    else 1
                )
                # self.debug(f"{e} exchangeable {int_partition}")
                self.log.detail(al, f"{e} exchangeable fix: {int_partition}")
                valid.append((e * prev_e, prop_vars))
        return valid

    def fix_non_exchangeable_partititons(self, rv_set, prev_e=1, var_list=None):
        # self.debug(f"Fixing {rv_set} entities on non-exchangeable partitions:")
        self.log.action(f"Fixing {rv_set} entities on non-exchangeable partitions")
        vars = var_list if var_list is not None else self.vars
        ex_classes = self.exchangeable_classes(vars)
        ics = self.feasible_relevant_compositions(ex_classes, rv_set)
        keys = list(ex_classes.keys())
        dists = []
        for distribution in ics:
            prop_vars = []
            # self.debug(
            #     f"Distribution of {rv_set} for {len(ex_classes)} classes: {distribution}"
            # )
            self.log.action(
                f"Distribution of {rv_set} for {len(ex_classes)} classes: {distribution}"
            )
            for j, n in enumerate(distribution):
                class_vars = ex_classes[keys[j]]
                prop_cvars = [v.copy() for v in class_vars]
                valid = self.fix_exchangeable_partititons(rv_set, n, 1, prop_cvars)
                if len(valid) > 0:
                    prop_vars.append(valid)
            if len(prop_vars) == len(ex_classes):  # otherwise something was unsat
                cases = list(self.product(prev_e, *prop_vars))
                dists += cases
        return dists

    def product(self, prev_e, *args):
        result = [(prev_e, [])]
        for pool in args:
            result = [(ex * ey, x + y) for ex, x in result for ey, y in pool]
        for prod in result:
            yield prod

    def fix_partitions(self, fixed, relevant):
        # fixed::(exchangeability constant, vars)
        if len(relevant) == 0:
            return [fixed]  # no more choices to account for
        else:
            al = self.log.action(f"Fixing {relevant} relevant sets")
            props = []
            e, vars = fixed
            rv_set = relevant[0]  # propagate one
            ex_classes = self.exchangeable_classes(vars)
            # self.debug(f"Propagating {rv_set} for {e}")
            self.log.detail(al, f"Propagating {rv_set} for set {e}")
            if len(ex_classes) == 1:
                prop_vars = self.fix_exchangeable_partititons(
                    rv_set, rv_set.size(), e, vars
                )
            else:
                prop_vars = self.fix_non_exchangeable_partititons(rv_set, e, vars)
            # self.debug(f"{rv_set} propagated")
            for pvs in prop_vars:
                prop_fix = self.fix_partitions(
                    pvs, relevant[1:]
                )  # propagate other relevants
                props += prop_fix
        return props

    def count_unfixed_partitions(self, relevant, var_list=None, log=None):
        vars = var_list if var_list is not None else self.vars
        log = self.log if log is None else log
        fixed = self.fix_partitions((1, vars), relevant)
        # self.debug(f"There are {len(fixed)} possible ways of fixing {relevant} domains")
        al = self.log.action("Counting unfixed partitions")
        self.log.detail(
            al, f"There are {len(fixed)} possible ways of fixing {relevant} domains"
        )
        count = Zero()
        for i, fix_vars in enumerate(fixed):
            e, prop_vars = fix_vars

            c_log = ProblemLog(
                prop_vars,
                universe=self.universe,
                caption=f"Fixing histograms, combination nr. {i+1}",
                id=i + 1,
                level=self.lvl + 1,
                config=self.config,
                debug=self.debug,
            )

            c_fix = self.count_fixed_partitions(
                relevant,
                prop_vars,
                caption=f"Fixed {i+1}",
                id=f"{self.log.id}.{str(i + 1)}",
                log=c_log,
            )
            c_log.count = c_fix.copy()
            log.add_subproblem("add", c_log)
            count += c_fix.with_choices(e)
        return count

    def feasible_relevant_partitions(self, rv_set, n, var_list=None):
        """
        Given a list of exchangerable variables LiftedSet, partitions n elements from rv_set.
        checking if there is some unsatisfiable constraint
        """
        vars = var_list if var_list is not None else self.vars
        int_partitions = self.integer_k_partitions(n, len(vars))
        fips = []
        al = self.log.action("Finding feasible relevant partitions")
        for int_partition in int_partitions:
            feasible = True
            for i, v in enumerate(vars):
                feasible = feasible and v.feasible(rv_set, int_partition[i])
            if feasible:
                fips.append(int_partition)
        # self.debug(f"{n} of {rv_set} have {len(fips)} feasible {len(vars)}-partitions")
        self.log.detail(
            al, f"{n} of {rv_set} have {len(fips)} feasible {len(vars)}-partitions"
        )
        return fips

    def feasible_relevant_compositions(self, ex_classes, rv_set):
        k = len(ex_classes)
        n = rv_set.size()
        int_partitions = self.integer_k_partitions(n, k)
        int_compositions = [
            list(c) for ip in int_partitions for c in set(itertools.permutations(ip))
        ]
        fics = []
        al = self.log.action("Finding feasible relevant compositions")
        for int_composition in int_compositions:
            feasible = True
            for i, v in enumerate(ex_classes.keys()):
                if feasible:
                    n_class = len(ex_classes[v])
                    f = v.feasible(rv_set, int_composition[i], n_class)
                    feasible = feasible and f
            if feasible:
                fics.append(int_composition)
        self.log.detail(
            al, f"{n} of {rv_set} have {len(fics)} feasible {k}-compositions"
        )
        return fics

    def n_int_perms(self, int_partition):
        # given the int partition returns the n!/n1!*n2!*...*nk! different permutations
        hst = self.histogram(int_partition)
        n = math.factorial(len(int_partition))
        p_classes = math.prod([math.factorial(hst[c]) for c in hst])
        return n // p_classes

    def propagate_ccs_partitions(self, vars, ccs, choices=1):
        """
        Given a list/set of variable partitions propagates counting constraints over partitions.
        If variables are exchangeable propagate, else shatter
        """
        # self.debug("Propagating size constraints...")
        al = self.log.action("Propagating size constraints")
        ex_classes = self.exchangeable_classes(vars)
        if len(ex_classes) == 1:  # pick a constraint and propagate
            cc = ccs[0]
            # self.debug(f"Variables are exchangeable: propagating {cc}")
            self.log.detail(al, f"Variables are exchangeable: propagating {cc}")
            if cc.values.upper == Int.inf:
                cc.values = cc.values.replace(upper=len(vars), right=Int.CLOSED)
            cases = []
            al2 = self.log.action("Shattering interval into single values")
            for i in Int.iterate(cc.values, step=1):
                self.log.detail(al2, f"Case  with {i} {cc.formula}")
                i_vars = vars.copy()
                cc_eq = CCounting(cc.formula, Int.singleton(i))
                try:
                    c_prop, p_vars = self.propagate_cc(
                        cc_eq, i_vars
                    )  # propagate one constraint
                    if len(ccs[1:]) > 0:
                        # self.debug("Shattering remaining...")
                        self.log.detail(al2, "Shattering remaining constraints")
                        shatter = self.propagate_ccs_partitions(
                            p_vars, ccs[1:], c_prop
                        )  # shatter others
                        cases += shatter
                    else:
                        cases += [(c_prop, p_vars)]
                except Unsatisfiable:
                    # self.debug("Propagation unsat")
                    self.log.detail(al2, "Propagation unsat")
            return cases
        else:  # shatter constraints
            # self.debug(f"Variables are not exchangeable: shattering size constraints")
            self.log.detail(
                al, f"Variables are not exchangeable: shattering size constraints"
            )
            scv, rcv = self.split_ex_classes(ex_classes)
            combs_split_class, combs_rest_classes = self.shatter_count_formulas(
                scv, rcv, ccs
            )
            vars_comb = []
            for i in range(0, len(combs_split_class)):
                cp_scv = scv.copy()
                cp_rcv = rcv.copy()
                prop_split_vars = self.propagate_ccs_partitions(
                    cp_scv, combs_split_class[i]
                )
                if len(prop_split_vars) == 1 and prop_split_vars[0][0] == 0:  # unsat
                    continue
                prop_rest_vars = self.propagate_ccs_partitions(
                    cp_rcv, combs_rest_classes[i]
                )
                if len(prop_rest_vars) == 1 and prop_rest_vars[0][0] == 0:  # unsat
                    continue
                ccs_combs = [
                    list(comb)
                    for comb in itertools.product(prop_split_vars, prop_rest_vars)
                ]
                for cc in ccs_combs:
                    s_problem, r_problem = cc
                    s_choices, svars = s_problem
                    r_choices, rvars = r_problem
                    c = choices * s_choices * r_choices
                    vs = svars + rvars
                    self.log.detail(al, f"Shattered as {vs}")
                    vars_comb.append((c, vs))
            return vars_comb

    def shatter_partitions(self):
        """
        Shatters partitions and combinations
        """
        if len(self.count_f) > 0:
            problems = self.propagate_ccs_partitions(self.vars, self.count_f)
        else:
            problems = [(1, vars)]
        count = Zero()
        # self.debug(
        #     f"Size constraints shattered, solving {len(problems)} combinations..."
        # )
        i = 1
        for c, p in problems:
            # self.debug(f"------ Combination {i} ------")
            i += 1
            prob_sol = self.count_partitions(p, i)
            self.log.add_subproblem("add", prob_sol.log)
            count += prob_sol.count.with_choices(c)
        return count

    def has_fixed_indist(self, var_list=None):
        vars = var_list if var_list is not None else self.vars
        ccs = vars[0].ccs
        indist_subsets = self.universe.indistinguishable_subsets()
        # indist = len(indist_subsets) > 0
        fixed = True
        for i in indist_subsets:
            fixed_set = False
            for cc in ccs:  # if a cc formula is a subset of i then i is partitioned
                fixed_set = fixed_set or (cc.formula in i)
            fixed = fixed and fixed_set  # if all sets are fixed then ok
        return fixed



import math


class Solution(object):
    def __init__(self, count, log):
        self.count = count
        self.log = log
        self.log.count = count.copy()

    def __str__(self):
        return str(self.count)


class Count(object):
    """
    A solution is a set of pairs (count, histogram).
    An histogram is used to keep track of used variables for injectivity constraint.
    For each histogram keep track of the corresponding number of (partial) solutions.
    """

    def __init__(self, count, histogram={}, subproblems=1, symbolic=None, tip=""):
        """
        A number plus a bunch of extra info to figure out where it comes from

        Args:
            count (int): the number
            histogram (dict, optional): A count for each relevant set used to derive this count. Defaults to {}.
            subproblems (int, optional): Counts the number of subproblems used to derive this count. Defaults to 1.
            symbolic (_type_, optional): A latex representation to display nicely the formula behind this count. Defaults to None.
            tip (str, optional): A text description of the symbolic formula. Defaults to "".
        """
        self.val = count
        self.histograms = [[count, histogram]]
        self.subproblems = subproblems
        self.symbolic = symbolic if symbolic is not None else str(count)
        self.tip = tip

    def __add__(self, rhs):
        self.val += rhs.val
        self.histograms += rhs.histograms
        self.subproblems += rhs.subproblems
        self.symbolic = self.latex_op("+", rhs)
        self.tip = ""
        return self

    def __mul__(self, rhs):
        histograms = []
        for c1, h1 in self.histograms:
            for c2, h2 in rhs.histograms:
                count = c1 * c2
                histogram = self.add_histograms(h1, h2)
                histograms.append([count, histogram])
        count = self.val * rhs.val
        subproblems = self.subproblems + rhs.subproblems
        symbolic = self.latex_op("\\cdot", rhs)
        mul = Count(count, subproblems=subproblems, symbolic=symbolic)
        mul.histograms = histograms
        return mul

    def __sub__(self, rhs):
        if rhs.val == 0:
            return self
        else:
            self.val -= rhs.val
            # self.histograms -= rhs.histograms
            self.subproblems += rhs.subproblems
            self.symbolic = self.latex_op("-", rhs)
            return self

    def __pow__(self, rhs):
        if rhs.val == 1:
            return self
        else:
            self.val = self.val**rhs.val
            self.subproblems += rhs.subproblems
            self.symbolic = self.latex_op("^", rhs)
            return self

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"{self.val} ({self.subproblems} subproblems)"

    def add_histograms(self, h1, h2):
        sum = {}
        for f1 in h1:
            if f1 in h2:
                sum[f1] = h1[f1] + h2[f1]
            else:
                sum[f1] = h1[f1]
        for f2 in h2:
            if f2 not in h1:
                sum[f2] = h2[f2]
        return sum

    def set_histogram(self, hst):
        self.histograms = [[self.val, hst]]

    def with_choices(self, n_choices, tip="Exchangeable choices"):
        updated_hst = []
        for c, hst in self.histograms:
            c = c * n_choices
            updated_hst.append([c, hst])
        self.val *= n_choices
        self.histograms = updated_hst
        self.symbolic = f"{self.add_tip(n_choices, tip)} \cdot {self.symbolic}"
        return self

    def latex(self):
        ignore = self.symbolic == ""
        no_tip = self.tip == ""
        if ignore:
            return ""
        else:
            if no_tip:
                return self.symbolic
            else:
                return self.add_tip(self.symbolic, self.tip)

    def latex_op(self, op, right):
        if self.latex() == "":
            return right.latex()
        elif right.latex() == "":
            return self.latex()
        else:
            return f"{self.latex()} {op} {right.latex()}"

    def add_tip(self, n, tip):
        return f"\\texttip{{ {n} }}{{ {tip} }}"

    def is_zero(self):
        return self.val == 0

    def copy(self):
        return Count(
            self.val, self.histograms, self.subproblems, self.symbolic, self.tip
        )


class Zero(Count):
    def __init__(self, histogram={}, tip="No solution found"):
        super().__init__(0, histogram=histogram, subproblems=0, symbolic="0", tip=tip)

    def __add__(self, rhs):
        return rhs


class One(Count):
    def __init__(self, histogram={}):
        super().__init__(1, histogram=histogram, subproblems=0, symbolic="")


class Binomial(Count):
    def __init__(self, n, m, tip=""):
        count = math.comb(n, m)
        symbolic = f"\\binom{{ {n} }}{{ {m} }}"
        super().__init__(count, subproblems=1, symbolic=symbolic, tip=tip)


class Stirling(Count):
    def __init__(self, n, m, tip=""):
        count = self.stirling(n, m)
        symbolic = f"\\genfrac{{\\{{}}{{\\}}}}{{0pt}}{{}}{{ {n} }}{{ {m} }}"
        super().__init__(count, subproblems=1, symbolic=symbolic, tip=tip)

    def stirling(self, n, k):
        computed = {}

        def stirling_aux(n, k):
            key = str(n) + "," + str(k)
            if key in computed.keys():
                return computed[key]
            if n == k == 0:
                return 1
            if (n > 0 and k == 0) or (n == 0 and k > 0):
                return 0
            if n == k:
                return 1
            if k > n:
                return 0
            result = k * stirling_aux(n - 1, k) + stirling_aux(n - 1, k - 1)
            computed[key] = result
            return result

        return stirling_aux(n, k)


class Divide(Count):
    def __init__(self, count_n, count_m, tip="", histogram={}):
        count = count_n.val // count_m.val
        subproblems = count_n.subproblems + count_m.subproblems
        symbolic = f"\\frac{{{count_n.latex()}}}{{{count_m.latex()}}}"
        super().__init__(
            count,
            histogram=histogram,
            subproblems=subproblems,
            symbolic=symbolic,
            tip=tip,
        )


class Factorial(Count):
    def __init__(self, n, tip=""):
        count = math.factorial(n)
        symbolic = f"{n}!"
        super().__init__(count, subproblems=1, symbolic=symbolic, tip=tip)



import portion as P

from .configuration import *
from .count import Zero, Solution
from .sharpCSP import SharpCSP
from .level_2 import LiftedSet
from .logger import ProblemLog


class Solver(object):
    """
    Sets up the decision variables from the problem
    """

    def __init__(self, problem, debug=False):
        self.debug = debug
        self.problem = problem
        self.universe = problem.universe
        n = self.universe.size()
        problem.configuration.size.update_upper_bound(n)
        for c in self.problem.constraints:
            c.update_upper_bound(n)
        self.size = problem.configuration.size
        self.config = problem.configuration
        self.log = ProblemLog(
            universe=self.universe,
            debug=debug,
            config=self.config,
            pos_constraints=self.problem.pos_constraints,
            constraints=problem.constraints,
            configuration=problem.configuration,
        )
        self.log.description("Root problem")

    def solve(self):
        unsat, msg = self.trivial_unsat()
        if not unsat:
            count = Zero()
            for n in self.size:
                if self.config.lvl1():
                    vars = [self.universe] * n
                else:
                    ub_size = self.universe.size() - n + 1
                    size = CSize("universe", P.closed(1, ub_size))
                    vars = [LiftedSet(self.universe, size)] * n
                csp = SharpCSP(
                    vars,
                    self.config,
                    self.problem.pos_constraints,
                    self.problem.constraints,
                    self.universe,
                    caption=f"Configuration of size {n}",
                    lvl=1,
                    debug=self.debug,
                )
                size_sol = csp.solve()

                count += size_sol.count
                self.log.add_subproblem("add", size_sol.log)
        else:
            count = Zero(tip=msg)

        return Solution(count, self.log)

    def trivial_unsat(self):
        """
        Check if there is some constraint unsatisfiable because of domains, regardless of the number of variables

        Args:
            log (bool, optional): Logging. Defaults to True.
        """
        if self.config.lvl1():
            for count_constr in self.problem.constraints:
                lower_b = count_constr.values.lower
                atleast = (
                    lower_b if count_constr.values.left == P.CLOSED else lower_b + 1
                )
                available_elems = count_constr.formula.size()
                if available_elems < atleast:
                    msg = f"Trivially unsat because we want at least {atleast} {count_constr.formula} but we have only {available_elems}"
                    self.log.description = msg
                    return (True, msg)
            return (False, "")
        else:
            # TODO
            return (False, "")



import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='coso',
    version='1.0.1',    
    url='https://github.com/PietroTotis/CoSo',
    author='Pietro Totis',
    author_email='pietro.totis@kuleuven.be',
    license='GNU General Public License (GPL)',
    packages=['src'],
    install_requires=["portion",
                      "clingo==5.5.1",
                      "psutil",
                      "yattag",
                      "ply"],    
    description = ("A solver for combinatorics math word problems."),
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: GNU General Public License (GPL)",
    ],
    keywords="combinatorics automated reasoning",
)